from __future__ import annotations

from typing import Callable, cast

import numpy as np
import pyqtgraph as pg
import qtpy
from pyqtgraph.GraphicsScene.mouseEvents import (
    HoverEvent as pgHoverEvent,
)
from pyqtgraph.GraphicsScene.mouseEvents import (
    MouseClickEvent as pgMouseClickEvent,
)
from qtpy import QtCore, QtGui
from qtpy.QtCore import Signal

from whitecanvas import protocols
from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._labels import Axis, AxisLabel, Ticks, Title
from whitecanvas.backend.pyqtgraph._qt_utils import from_qt_button, from_qt_modifiers
from whitecanvas.types import MouseButton, MouseEvent, MouseEventType


@protocols.check_protocol(protocols.CanvasProtocol)
class Canvas:
    def __init__(
        self,
        item: pg.PlotItem | None = None,
        *,
        xaxis: str = "bottom",
        yaxis: str = "left",
    ):
        # prepare widget
        if item is None:
            viewbox = pg.ViewBox()
            item = pg.PlotItem(viewBox=viewbox)
        item.vb.disableAutoRange()  # auto range is done in the whitecanvas side
        self._signals = SignalListener()
        item.vb.sigRangeChanged.connect(self._signals._set_rect)
        item.addItem(self._signals)
        self._plot_item = item
        self._xaxis = Axis(self, axis=xaxis)
        self._yaxis = Axis(self, axis=yaxis)
        self._xticks = Ticks(self, axis=xaxis)
        self._yticks = Ticks(self, axis=yaxis)
        self._title = Title(self)
        self._xlabel = AxisLabel(self, axis=xaxis)
        self._ylabel = AxisLabel(self, axis=yaxis)
        self._last_event: MouseEvent = None

    def _plt_get_native(self):
        return self._plot_item

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_reorder_layers(self, layers: list[PyQtLayer]):
        for i, layer in enumerate(layers):
            layer._plt_set_zorder(i)

    def _plt_get_aspect_ratio(self) -> float | None:
        """Get aspect ratio of canvas"""
        locked = self._viewbox().state["aspectLocked"]
        if not locked:
            return None
        return float(locked)

    def _plt_set_aspect_ratio(self, ratio: float | None):
        """Set aspect ratio of canvas"""
        if ratio is None:
            self._viewbox().setAspectLocked(lock=False)
        else:
            self._viewbox().setAspectLocked(lock=True, ratio=ratio)

    def _viewbox(self) -> pg.ViewBox:
        return self._plot_item.vb

    def _plt_add_layer(self, layer: protocols.BaseProtocol):
        self._plot_item.addItem(layer)

    def _plt_remove_layer(self, layer):
        """Remove layer from the canvas"""
        self._plot_item.removeItem(layer)

    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""
        return self._plot_item.isVisible()

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""
        self._plot_item.setVisible(visible)

    def _plt_twinx(self) -> Canvas:
        """Create a twinx canvas"""
        plotitem = self._plot_item
        vb1 = plotitem.vb
        vb2 = pg.ViewBox()
        canvas = Canvas(pg.PlotItem(viewBox=vb2), yaxis="right")

        self._get_scene().addItem(vb2)
        plotitem.getAxis("right").linkToView(vb2)
        vb2.setXLink(plotitem)

        def _update_views():
            vb2.setGeometry(vb1.sceneBoundingRect())
            vb2.linkedViewChanged(vb1, vb2.XAxis)

        _update_views()
        vb1.sigResized.connect(_update_views)

        return canvas

    def _plt_twiny(self) -> Canvas:
        plotitem = self._plot_item
        vb1 = plotitem.vb
        vb2 = pg.ViewBox()
        canvas = Canvas(pg.PlotItem(viewBox=vb2), xaxis="top")

        self._get_scene().addItem(vb2)
        plotitem.getAxis("bottom").linkToView(vb2)
        vb2.setYLink(plotitem)

        def _update_views():
            vb2.setGeometry(vb1.sceneBoundingRect())
            vb2.linkedViewChanged(vb1, vb2.YAxis)

        _update_views()
        vb1.sigResized.connect(_update_views)

        return canvas

    # def _plt_inset(self, rect: Rect) -> Canvas:
    #     ...

    def _get_scene(self) -> pg.GraphicsScene:
        return self._plot_item.scene()

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev):
            mev = self._translate_mouse_event(ev, MouseEventType.CLICK)
            callback(mev)

        self._signals.pressed.connect(_cb)

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(qpoint: QtCore.QPointF):
            mev = self._translate_mouse_event(qpoint, MouseEventType.MOVE)
            callback(mev)

        self._signals.moved.connect(_cb)

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev):
            mev = self._translate_mouse_event(ev, MouseEventType.DOUBLE_CLICK)
            callback(mev)

        self._signals.double_clicked.connect(_cb)

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev):
            mev = self._translate_mouse_event(ev, MouseEventType.RELEASE)
            callback(mev)

        self._signals.released.connect(_cb)

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to x-limits changed event"""
        self._plot_item.sigXRangeChanged.connect(lambda _, x: callback(x))

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to y-limits changed event"""
        self._plot_item.sigYRangeChanged.connect(lambda _, y: callback(y))

    def _plt_draw(self):
        pass  # pyqtgraph has its own draw mechanism

    def _translate_mouse_event(
        self,
        ev: QtGui.QMouseEvent | QtCore.QPointF,
        typ: MouseEventType,
    ) -> MouseEvent:
        """Translate a mouse event from pyqtgraph to whitecanvas."""
        if isinstance(ev, QtCore.QPointF):
            # Hover events only have positions.
            qpoint = ev
            scene = self._get_scene()
            btns = scene.dragButtons
            evs: list[pgMouseClickEvent] = scene.clickEvents
            if len(btns) == 0:
                button = MouseButton.NONE
            else:
                button = from_qt_button(btns[0])
            if len(evs) == 0:
                modifiers = ()
            else:
                modifiers = from_qt_modifiers(evs[0].modifiers())
            mev = MouseEvent(
                button=button,
                modifiers=modifiers,
                pos=(qpoint.x(), qpoint.y()),
                type=typ,
            )
        else:
            qpoint = cast(QtCore.QPointF, ev.pos())

            modifiers = from_qt_modifiers(ev.modifiers())

            mev = MouseEvent(
                button=from_qt_button(ev.button()),
                modifiers=modifiers,
                pos=(qpoint.x(), qpoint.y()),
                type=typ,
            )
        return mev


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[int], widths: list[int], app: str = "default"):
        if app == "notebook":
            from pyqtgraph.jupyter import GraphicsLayoutWidget
        elif app in ("default", "qt"):
            from pyqtgraph import GraphicsLayoutWidget
        else:
            raise ValueError(f"pyqtgraph does not support {app!r}")
        self._layoutwidget = GraphicsLayoutWidget()
        self._heights = heights  # TODO: not used
        self._widths = widths

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int) -> Canvas:
        vb = pg.ViewBox()
        item = pg.PlotItem(viewBox=vb)
        self._layoutwidget.addItem(item, row, col)
        if rowspan != 1:
            self._layoutwidget.ci.layout.setRowStretchFactor(row, rowspan)
        if colspan != 1:
            self._layoutwidget.ci.layout.setColumnStretchFactor(col, colspan)
        canvas = Canvas(item)
        return canvas

    def _plt_show(self) -> bool:
        self._layoutwidget.setVisible(True)

    def _get_background_brush(self) -> QtGui.QBrush:
        return self._layoutwidget.backgroundBrush()

    def _plt_get_background_color(self):
        brush = self._get_background_brush()
        return np.array(brush.color().getRgbF())

    def _plt_set_background_color(self, color):
        brush = self._get_background_brush()
        brush.setColor(QtGui.QColor.fromRgbF(*color))
        self._layoutwidget.setBackgroundBrush(brush)

    def _plt_screenshot(self):
        img: QtGui.QImage = self._layoutwidget.grab().toImage()
        bits = img.constBits()
        h, w, c = img.height(), img.width(), 4
        if qtpy.API_NAME.startswith("PySide"):
            arr = np.asarray(bits).reshape(h, w, c)
        else:
            bits.setsize(h * w * c)
            arr = np.frombuffer(bits, np.uint8).reshape(h, w, c)

        return arr[:, :, [2, 1, 0, 3]]

    def _plt_set_figsize(self, width: int, height: int):
        self._layoutwidget.resize(width, height)


class SignalListener(pg.GraphicsObject):
    # Mouse events in pyqtgraph is very complicated.
    # Adding this graphics object to the scene will make it easier.
    pressed = Signal(object)
    moved = Signal(object)
    released = Signal(object)
    double_clicked = Signal(object)

    def __init__(self):
        super().__init__()
        self._rect = QtCore.QRectF(-10000, -10000, 10000, 10000)
        self._had_button = False

    def _set_rect(self, vb, rng: tuple[tuple[float, float], tuple[float, float]], _):
        (x0, x1), (y0, y1) = rng
        self._rect = QtCore.QRectF(x0, y0, x1 - x0, y1 - y0)

    def boundingRect(self):
        return self._rect

    def mousePressEvent(self, ev: pgMouseClickEvent):
        self.pressed.emit(ev)
        self._had_button = bool(ev.buttons() ^ QtCore.Qt.MouseButton.NoButton)
        super().mousePressEvent(ev)

    def hoverEvent(self, ev: pgHoverEvent):
        if ev.isExit():
            return
        btns: QtCore.Qt.MouseButtons = ev.buttons()

        has_button = bool(btns ^ QtCore.Qt.MouseButton.NoButton)
        if (not has_button) and self._had_button:
            self.released.emit(ev.pos())
        else:
            self.moved.emit(ev.pos())
        self._had_button = has_button

    def mouseDoubleClickEvent(self, ev):
        self.double_clicked.emit(ev)
        super().mouseDoubleClickEvent(ev)

    def paint(self, *args):
        pass



import pytest
from whitecanvas import new_canvas
import numpy as np


def test_hist():
    canvas = new_canvas(backend="mock")
    rng = np.random.default_rng(1642)
    data = np.concatenate([rng.normal(size=50), rng.normal(1.2, size=40)])
    canvas.add_hist(data, bins="auto")
    canvas.add_hist(data, bins=10)
    canvas.add_hist(data, bins=10, limits=(-2, 2))
    canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="step")
    canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="polygon")
    canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="bars")
    with pytest.raises(ValueError):
        canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="not-a-shape")
    canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="density")
    canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="probability")
    canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="frequency")
    canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="percent")

def test_hover_template():
    canvas = new_canvas(backend="mock")
    layer = canvas.add_markers([1, 2, 3], [4, 5, 6])
    layer.with_hover_template("const-text")
    layer.with_hover_template("{x}")
    layer.with_hover_template("{x}, {y}")
    layer.with_hover_template("{i}")
    layer.with_hover_template("a: {a}", extra={"a": [0, 1, 4]})
    layer.with_hover_template("x: {x}", extra={"x": [0, 1, 4]})
    layer.with_hover_template("a: {a:.2f}", extra={"a": [0, 1, 4.0], "b": [1, 2, 3]})
    with pytest.raises(KeyError):
        layer.with_hover_template("c: {c}", extra={"a": [0, 1, 4], "b": [1, 2, 3]})
    with pytest.raises(ValueError):
        layer.with_hover_template("c: {c}", extra={"c": [0, 1, 4, 5]})



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend import _not_implemented
from whitecanvas.backend.plotly._base import (
    PlotlyLayer,
    from_plotly_marker_symbol,
    to_plotly_marker_symbol,
)
from whitecanvas.protocols import MarkersProtocol, check_protocol
from whitecanvas.types import Symbol
from whitecanvas.utils.normalize import arr_color, as_color_array, rgba_str_color
from whitecanvas.utils.type_check import is_real_number


@check_protocol(MarkersProtocol)
class Markers(PlotlyLayer):
    def __init__(self, xdata, ydata):
        ndata = len(xdata)
        self._props = {
            "x": xdata,
            "y": ydata,
            "mode": "markers",
            "marker": {
                "color": ["blue"] * ndata,
                "size": np.full(ndata, 10),
                "symbol": "circle",
                "line": {"width": np.ones(ndata), "color": ["blue"] * ndata},
            },
            "type": "scatter",
            "showlegend": False,
            "visible": True,
            "customdata": list(zip([""] * ndata, [id(self)] * ndata)),
            "hovertemplate": "%{customdata[0]}<extra></extra>",
        }
        self._fig_ref = lambda: None
        self._click_callbacks = []

    def _plt_get_ndata(self) -> int:
        return len(self._props["x"])

    def _plt_get_data(self):
        return self._props["x"], self._props["y"]

    def _plt_set_data(self, xdata, ydata):
        self._props["x"] = xdata
        self._props["y"] = ydata

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        color = self._props["marker"]["color"]
        if len(color) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([arr_color(c) for c in color], axis=0)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, self._plt_get_ndata())
        self._props["marker"]["color"] = [rgba_str_color(c) for c in color]

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_patterns()

    def _plt_get_symbol(self) -> Symbol:
        return from_plotly_marker_symbol(self._props["marker"]["symbol"])

    def _plt_set_symbol(self, symbol: Symbol):
        self._props["marker"]["symbol"] = to_plotly_marker_symbol(symbol)

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        return np.asarray(self._props["marker"]["size"])

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(self._plt_get_ndata(), size)
        self._props["marker"]["size"] = size

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return np.asarray(self._props["marker"]["line"]["width"])

    def _plt_set_edge_width(self, width: float):
        if is_real_number(width):
            width = np.full(self._plt_get_ndata(), width)
        self._props["marker"]["line"]["width"] = width

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_styles()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        color = self._props["marker"]["line"]["color"]
        if len(color) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([arr_color(c) for c in color], axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, self._plt_get_ndata())
        self._props["marker"]["line"]["color"] = [rgba_str_color(c) for c in color]

    def _plt_connect_pick_event(self, callback):
        fig = self._fig_ref()
        if fig is None:
            self._click_callbacks.append(callback)
            return
        else:
            raise NotImplementedError("post connection not implemented yet")

    def _plt_set_hover_text(self, text: list[str]):
        fig = self._fig_ref()
        if fig is None:
            return

        # check by ID
        def selector(trace):
            s = trace["customdata"]
            if s is None:
                return False
            return s[0][1] == id(self)

        fig.update_traces(
            customdata=list(zip(text, [id(self)] * self._plt_get_ndata())),
            selector=selector,
        )



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from whitecanvas import new_canvas


def test_line_marker_text(backend: str):
    canvas = new_canvas(backend=backend)
    layer = canvas.add_line(
        np.arange(10), np.arange(10) * 2
    ).with_markers(
        symbol="s", size=10, color="red"
    ).with_text(
        [f"Data-{i}" for i in range(10)], color="blue"
    ).with_yerr(
        np.ones(10), color="blue"
    )
    assert_array_equal(layer.data.x, np.arange(10))
    assert_array_equal(layer.data.y, np.arange(10) * 2)
    assert len(canvas.layers) == 1


def test_bar_err_text(backend: str):
    canvas = new_canvas(backend=backend)
    layer = canvas.add_bars(
        np.arange(10), np.arange(10) * 2
    ).with_err(
        np.ones(10), color="blue"
    ).with_text(
        [f"Data-{i}" for i in range(10)], color="blue"
    )
    assert_array_equal(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10) * 2, atol=1e-6)
    assert len(canvas.layers) == 1

def test_stem(backend: str):
    canvas = new_canvas(backend=backend)
    layer = canvas.add_markers(
        np.arange(10), np.arange(10) * 2
    ).with_stem()
    assert_array_equal(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10) * 2, atol=1e-6)
    assert len(canvas.layers) == 1

def test_network(backend: str):
    canvas = new_canvas(backend=backend)
    layer = canvas.add_markers(
        np.arange(10), np.arange(10) * 2
    ).with_network([[0, 1], [0, 2], [1, 3]])
    assert_array_equal(layer.nodes.data.x, np.arange(10))
    assert_array_equal(layer.nodes.data.y, np.arange(10) * 2)
    assert len(canvas.layers) == 1



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar, overload

from numpy.typing import ArrayLike

from whitecanvas import layers as _l
from whitecanvas.canvas import Canvas
from whitecanvas.layers import _mixin
from whitecanvas.plot._canvases import current_canvas
from whitecanvas.types import (
    Alignment,
    ArrayLike1D,
    ColormapType,
    ColorType,
    Hatch,
    HistBinType,
    LineStyle,
    Orientation,
    Symbol,
)

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec

    from whitecanvas.canvas._stacked import StackOverPlotter

    _P = ParamSpec("_P")

_F = TypeVar("_F")
_L = TypeVar("_L", bound=_l.Layer)


def _copy_method(f: _F) -> _F:
    """Search for the corresponding Canvas method and wrap."""
    return _copy_add_method(f, "")


def _copy_add_method(f: _F, pref: str = "add_") -> _F:
    """Search for the corresponding Canvas.add_* method and wrap."""
    fname = f"{pref}{f.__name__}"

    def _inner(*args, **kwargs):
        meth = getattr(current_canvas(), fname)
        return meth(*args, **kwargs)

    canvas_func = getattr(Canvas, fname)
    doc = canvas_func.__doc__
    assert isinstance(doc, str)
    _inner.__doc__ = doc.replace(f">>> canvas.{pref}", ">>> plt.")
    _inner.__name__ = f.__name__
    _inner.__qualname__ = f.__qualname__
    _inner.__module__ = f.__module__
    _inner.__annotations__ = canvas_func.__annotations__
    return _inner


@overload
def line(
    ydata: ArrayLike1D, *, name: str | None = None, color: ColorType | None = None,
    width: float = 1.0, style: LineStyle | str = LineStyle.SOLID, alpha: float = 1.0,
    antialias: bool = True,
) -> _l.Line:  # fmt: skip
    ...


@overload
def line(
    xdata: ArrayLike1D, ydata: ArrayLike1D, *, name: str | None = None,
    color: ColorType | None = None, width: float = 1.0,
    style: LineStyle | str = LineStyle.SOLID, alpha: float = 1.0,
    antialias: bool = True,
) -> _l.Line:  # fmt: skip
    ...


@overload
def line(
    xdata: ArrayLike1D, ydata: Callable[[ArrayLike1D], ArrayLike1D], *,
    name: str | None = None, color: ColorType | None = None, width: float = 1.0,
    style: LineStyle | str = LineStyle.SOLID, alpha: float = 1.0,
    antialias: bool = True,
) -> _l.Line:  # fmt: skip
    ...


@_copy_add_method
def line(*args, **kwargs):
    ...


@overload
def markers(
    xdata: ArrayLike1D, ydata: ArrayLike1D, *,
    name: str | None = None, symbol: Symbol | str = Symbol.CIRCLE,
    size: float = 12, color: ColorType | None = None, alpha: float = 1.0,
    hatch: str | Hatch = Hatch.SOLID,
) -> _l.Markers[_mixin.ConstFace, _mixin.ConstEdge, float]:  # fmt: skip
    ...


@overload
def markers(
    ydata: ArrayLike1D, *,
    name: str | None = None, symbol: Symbol | str = Symbol.CIRCLE,
    size: float = 12, color: ColorType | None = None, alpha: float = 1.0,
    hatch: str | Hatch = Hatch.SOLID,
) -> _l.Markers[_mixin.ConstFace, _mixin.ConstEdge, float]:  # fmt: skip
    ...


@_copy_add_method
def markers(*args, **kwargs):
    ...


@overload
def bars(
    center: ArrayLike1D, height: ArrayLike1D, *, bottom: ArrayLike1D | None = None,
    name=None, orient: str | Orientation = Orientation.VERTICAL,
    extent: float = 0.8, color: ColorType | None = None,
    alpha: float = 1.0, hatch: str | Hatch = Hatch.SOLID,
) -> _l.Bars[_mixin.ConstFace, _mixin.ConstEdge]:  # fmt: skip
    ...


@overload
def bars(
    height: ArrayLike1D, *, bottom: ArrayLike1D | None = None,
    name=None, orient: str | Orientation = Orientation.VERTICAL,
    extent: float = 0.8, color: ColorType | None = None,
    alpha: float = 1.0, hatch: str | Hatch = Hatch.SOLID,
) -> _l.Bars[_mixin.ConstFace, _mixin.ConstEdge]:  # fmt: skip
    ...


@_copy_add_method
def bars(*args, **kwargs):
    ...


@_copy_add_method
def band(
    xdata: ArrayLike1D,
    ylow: ArrayLike1D,
    yhigh: ArrayLike1D,
    *,
    name: str | None = None,
    orient: str | Orientation = Orientation.VERTICAL,
    color: ColorType | None = None,
    alpha: float = 1.0,
    hatch: str | Hatch = Hatch.SOLID,
) -> _l.Band:
    ...


@_copy_add_method
def infline(
    pos: tuple[float, float] = (0, 0),
    angle: float = 0.0,
    *,
    name: str | None = None,
    color: ColorType | None = None,
    width: float = 1.0,
    style: LineStyle | str = LineStyle.SOLID,
    alpha: float = 1.0,
    antialias: bool = True,
) -> _l.InfLine:
    ...


@_copy_add_method
def infcurve(
    model: Callable[Concatenate[Any, _P], Any],
    *,
    bounds: tuple[float, float] = (-float("inf"), float("inf")),
    name: str | None = None,
    color: ColorType | None = None,
    width: float = 1.0,
    style: str | LineStyle = LineStyle.SOLID,
    antialias: bool = True,
) -> _l.InfCurve[_P]:
    ...


@_copy_add_method
def errorbars(
    xdata: ArrayLike1D,
    ylow: ArrayLike1D,
    yhigh: ArrayLike1D,
    *,
    name: str | None = None,
    orient: str | Orientation = Orientation.VERTICAL,
    color: ColorType = "blue",
    width: float = 1,
    style: LineStyle | str = LineStyle.SOLID,
    antialias: bool = False,
    capsize: float = 0.0,
) -> _l.Errorbars:
    ...


@_copy_add_method
def hist(
    data: ArrayLike1D,
    *,
    bins: HistBinType = "auto",
    range: tuple[float, float] | None = None,
    density: bool = False,
    name: str | None = None,
    orient: str | Orientation = Orientation.VERTICAL,
    color: ColorType | None = None,
    alpha: float = 1.0,
    hatch: str | Hatch = Hatch.SOLID,
) -> _l.Bars:
    ...


@_copy_add_method
def rug(
    events: ArrayLike1D,
    *,
    low: float = 0.0,
    high: float = 1.0,
    name: str | None = None,
    orient: str | Orientation = Orientation.VERTICAL,
    color: ColorType = "black",
    width: float = 1.0,
    style: LineStyle | str = LineStyle.SOLID,
    antialias: bool = True,
    alpha: float = 1.0,
) -> _l.Rug:
    ...


@_copy_add_method
def spans(
    spans: ArrayLike,
    *,
    name: str | None = None,
    orient: str | Orientation = Orientation.VERTICAL,
    color: ColorType = "blue",
    alpha: float = 0.2,
    hatch: str | Hatch = Hatch.SOLID,
) -> _l.Spans:
    ...


@_copy_add_method
def image(
    image: ArrayLike,
    *,
    cmap: ColormapType = "gray",
    clim: tuple[float | None, float | None] | None = None,
    flip_canvas: bool = True,
    lock_aspect: bool = True,
) -> _l.Image:
    ...


@_copy_add_method
def text(
    x: ArrayLike1D,
    y: ArrayLike1D,
    string: list[str],
    *,
    color: ColorType = "black",
    size: float = 12,
    rotation: float = 0.0,
    anchor: str | Alignment = Alignment.BOTTOM_LEFT,
    family: str | None = None,
) -> _l.Texts[_mixin.ConstFace, _mixin.ConstEdge, _mixin.ConstFont]:
    ...


@_copy_add_method
def kde(
    data: ArrayLike1D,
    *,
    bottom: float = 0.0,
    name: str | None = None,
    orient: str | Orientation = Orientation.VERTICAL,
    band_width: float | Literal["scott", "silverman"] = "scott",
    color: ColorType | None = None,
    alpha: float = 1.0,
    hatch: str | Hatch = Hatch.SOLID,
) -> _l.Band:
    ...


@_copy_method
def cat(
    data: Any,
    by: str | None = None,
    *,
    orient: str | Orientation = Orientation.VERTICAL,
    offsets: float | ArrayLike1D | None = None,
    palette: ColormapType | None = None,
    update_labels: bool = True,
):
    ...


@_copy_method
def stack_over(
    layer: _L,
) -> StackOverPlotter[Canvas, _L]:
    ...



from __future__ import annotations

from typing import TYPE_CHECKING

from numpy.typing import ArrayLike

from whitecanvas.layers._primitive import Errorbars, Line, Markers, Texts
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.types import Alignment, ColorType, LineStyle, Orientation, _Void

if TYPE_CHECKING:
    from whitecanvas.layers.group.labeled import LabeledPlot

_void = _Void()


class Plot(LayerContainer):
    """A Plot layer is composed of a line and a markers layer."""

    def __init__(
        self,
        line: Line,
        markers: Markers,
        name: str | None = None,
    ):
        super().__init__([line, markers], name=name)

    @property
    def line(self) -> Line:
        """The line layer."""
        return self._children[0]

    @property
    def markers(self) -> Markers:
        """The markers layer."""
        return self._children[1]

    def with_edge(
        self,
        color: ColorType | _Void = _void,
        alpha: float = 1.0,
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Plot:
        self.markers.with_edge(color=color, alpha=alpha, width=width, style=style)
        return self

    def with_text(
        self,
        strings: list[str],
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        fontfamily: str | None = None,
    ) -> LabeledPlot:
        from whitecanvas.layers.group.labeled import LabeledPlot

        texts = Texts(
            *self.line.data,
            strings,
            color=color,
            size=size,
            rotation=rotation,
            anchor=anchor,
            family=fontfamily,
            backend=self._backend_name,
        )
        xerr = Errorbars.empty(Orientation.VERTICAL, backend=self._backend_name)
        yerr = Errorbars.empty(Orientation.HORIZONTAL, backend=self._backend_name)
        return LabeledPlot(self, xerr, yerr, texts, name=self.name)

    def with_xerr(
        self,
        err: ArrayLike,
        err_high: ArrayLike | None = None,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
        capsize: float = 0,
    ) -> LabeledPlot:
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group.labeled import LabeledPlot

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.line.color
        if width is _void:
            width = self.markers.edge.width
        if style is _void:
            style = self.markers.edge.style
        if antialias is _void:
            antialias = self.line.antialias
        xerr = Errorbars(
            self.data.y, self.data.x - err, self.data.x + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            backend=self._backend_name
        )  # fmt: skip
        yerr = Errorbars.empty(Orientation.HORIZONTAL, backend=self._backend_name)
        return LabeledPlot(self, xerr, yerr, name=self.name)

    def with_yerr(
        self,
        err: ArrayLike,
        err_high: ArrayLike | None = None,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
        capsize: float = 0,
    ) -> LabeledPlot:
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group.labeled import LabeledPlot

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.line.color
        if width is _void:
            width = self.markers.edge.width
        if style is _void:
            style = self.markers.edge.style
        if antialias is _void:
            antialias = self.line.antialias
        yerr = Errorbars(
            self.data.x, self.data.y - err, self.data.y + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            backend=self._backend_name
        )  # fmt: skip
        xerr = Errorbars.empty(Orientation.VERTICAL, backend=self._backend_name)
        return LabeledPlot(self, xerr, yerr, name=self.name)

    @property
    def data(self):
        """The internal data of the line and markers."""
        return self.line.data

    def set_data(self, xdata=None, ydata=None):
        self.line.set_data(xdata, ydata)
        self.markers.set_data(xdata, ydata)



import matplotlib.pyplot as plt
import pytest


@pytest.fixture(
    scope="function",
    params=["mock", "matplotlib", "pyqtgraph", "plotly", "bokeh", "vispy"]
)
def backend(request: pytest.FixtureRequest):
    yield request.param
    # TODO: how to skip tests if failed in mock backend?
    if request.param == "matplotlib":
        plt.close("all")



from __future__ import annotations

from whitecanvas.backend import Backend
from whitecanvas.canvas import Canvas, CanvasGrid
from whitecanvas.core import grid, new_canvas


def current_grid() -> CanvasGrid:
    """Return the current canvas grid."""
    grid = CanvasGrid._CURRENT_INSTANCE
    if grid is None:
        grid = new_canvas()._grid
    return grid


def current_canvas() -> Canvas:
    """Return the current canvas."""
    canvas = Canvas._CURRENT_INSTANCE
    if canvas is None:
        canvas = new_canvas()
    return canvas


def show(block: bool = False):
    """Show the current canvas."""
    current_grid().show(block=block)


def subplots(
    nrows: int = 1,
    ncols: int = 1,
    *,
    backend: Backend | str | None = None,
) -> CanvasGrid:
    """Create a new grid of subplots."""
    return grid(nrows, ncols, backend=backend).fill()



from __future__ import annotations

from typing import Callable

import numpy as np
from bokeh import events as bk_events
from bokeh import layouts as bk_layouts
from bokeh import models as bk_models
from bokeh import plotting as bk_plotting
from bokeh.io import output_notebook

from whitecanvas import protocols
from whitecanvas.backend.bokeh._base import BokehLayer
from whitecanvas.backend.bokeh._labels import (
    Title,
    XAxis,
    XLabel,
    XTicks,
    YAxis,
    YLabel,
    YTicks,
)
from whitecanvas.types import Modifier, MouseButton, MouseEvent, MouseEventType
from whitecanvas.utils.normalize import arr_color, hex_color


def _prep_plot(width=400, height=300):
    plot = bk_plotting.figure(
        width=width,
        height=height,
        tooltips="@hovertexts",
    )
    plot.title.align = "center"
    return plot


SECOND_Y = "second-y"


@protocols.check_protocol(protocols.CanvasProtocol)
class Canvas:
    def __init__(
        self,
        plot: bk_models.Plot | None = None,
        second_y: bool = False,
    ):
        if plot is None:
            plot = _prep_plot()
        assert isinstance(plot, bk_models.Plot)
        self._plot = plot
        self._xaxis = XAxis(self)
        self._yaxis = YAxis(self)
        self._title = Title(self)
        self._xlabel = XLabel(self)
        self._ylabel = YLabel(self)
        self._xticks = XTicks(self)
        self._yticks = YTicks(self)
        self._mouse_button: MouseButton = MouseButton.NONE
        self._second_y = second_y

    def _set_mouse_down(self, event):
        self._mouse_button = event

    def _get_xaxis(self):
        return self._plot.xaxis[0]

    def _get_yaxis(self):
        if not self._second_y:
            return self._plot.yaxis[0]
        return self._plot.yaxis[1]

    def _get_xrange(self):
        return self._plot.x_range

    def _get_yrange(self):
        if not self._second_y:
            return self._plot.y_range
        return self._plot.extra_y_ranges[SECOND_Y]

    def _plt_get_native(self):
        return self._plot

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _bokeh_renderers(self) -> list[bk_models.GlyphRenderer]:
        return self._plot.renderers

    def _plt_reorder_layers(self, layers: list[BokehLayer]):
        model_to_idx_map = {id(layer._model): i for i, layer in enumerate(layers)}
        renderes = self._bokeh_renderers()
        self._plot.renderers = [
            renderes[model_to_idx_map[id(r.glyph)]] for r in renderes
        ]

    def _plt_get_aspect_ratio(self) -> float | None:
        return self._plot.aspect_ratio

    def _plt_set_aspect_ratio(self, ratio: float | None):
        self._plot.aspect_ratio = ratio

    def _plt_add_layer(self, layer: BokehLayer):
        if self._second_y:
            self._plot.add_glyph(layer._data, layer._model, y_range_name=SECOND_Y)
        else:
            self._plot.add_glyph(layer._data, layer._model)

    def _plt_remove_layer(self, layer: BokehLayer):
        """Remove layer from the canvas"""
        idx = -1
        for i, renderer in enumerate(self._bokeh_renderers()):
            if renderer.glyph == layer._model:
                idx = i
                break
        del self._plot.renderers[idx]

    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""
        return self._plot.visible

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""
        self._plot.visible = visible

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.Press):
            self._set_mouse_down(MouseButton.LEFT)
            ev = MouseEvent(
                button=MouseButton.LEFT,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.CLICK,
            )
            callback(ev)

        self._plot.on_event(bk_events.Tap, _cb)
        self._plot.on_event(bk_events.Press, _cb)

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.MouseMove):
            ev = MouseEvent(
                button=self._mouse_button,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.MOVE,
            )
            callback(ev)

        self._plot.on_event(bk_events.MouseMove, _cb)

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.DoubleTap):
            ev = MouseEvent(
                button=MouseButton.LEFT,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.DOUBLE_CLICK,
            )
            callback(ev)

        self._plot.on_event(bk_events.DoubleTap, _cb)

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.PressUp):
            ev = MouseEvent(
                button=MouseButton.LEFT,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.RELEASE,
            )
            callback(ev)
            self._set_mouse_down(MouseButton.NONE)

        self._plot.on_event(bk_events.PressUp, _cb)
        self._plot.on_event(bk_events.Tap, _cb)
        self._plot.on_event(bk_events.PanEnd, _cb)

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        rng = self._plot.x_range
        rng.on_change(
            "start",
            lambda attr, old, new: callback((rng.start, rng.end)),  # noqa: ARG005
        )

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        rng = self._plot.y_range
        rng.on_change(
            "start",
            lambda attr, old, new: callback((rng.start, rng.end)),  # noqa: ARG005
        )

    def _plt_draw(self):
        pass

    def _plt_twinx(self):
        self._plot.add_layout(
            bk_models.LinearAxis(y_range_name=SECOND_Y),
            "right",
        )
        self._plot.extra_y_ranges = {SECOND_Y: bk_models.DataRange1d()}
        return Canvas(self._plot, second_y=True)


def _translate_modifiers(mod: bk_events.KeyModifiers | None) -> tuple[Modifier, ...]:
    if mod is None:
        return ()
    modifiers = []
    if mod["alt"]:
        modifiers.append(Modifier.ALT)
    if mod["shift"]:
        modifiers.append(Modifier.SHIFT)
    if mod["ctrl"]:
        modifiers.append(Modifier.CTRL)
    return tuple(modifiers)


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[int], widths: list[int], app: str = "default"):
        nr, nc = len(heights), len(widths)
        children = []
        for _ in range(nr):
            row = []
            for _ in range(nc):
                row.append(_prep_plot())
            children.append(row)
        self._grid_plot: bk_layouts.GridPlot = bk_layouts.gridplot(
            children, sizing_mode="fixed"
        )
        self._shape = (nr, nc)
        self._app = app

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int) -> Canvas:
        for plot, r0, c0 in self._grid_plot.children:
            if r0 == row and c0 == col:
                return Canvas(plot)
        raise ValueError(f"Canvas at ({row}, {col}) not found")

    def _plt_show(self):
        if self._app == "notebook":
            output_notebook()
            bk_plotting.show(lambda doc: doc.add_root(self._grid_plot))
        else:
            bk_plotting.show(self._grid_plot)
            if self._grid_plot.document is None:
                bk_plotting.curdoc().add_root(self._grid_plot)

    def _plt_get_background_color(self):
        return arr_color(self._grid_plot.background_fill_color)

    def _plt_set_background_color(self, color):
        color = hex_color(color)
        for r in range(self._shape[0]):
            for c in range(self._shape[1]):
                child = self._grid_plot.children[r][c]
                if not hasattr(child, "background_fill_color"):
                    continue
                child.background_fill_color = color

    def _plt_screenshot(self):
        import io

        from bokeh.io import export_png

        with io.BytesIO() as buff:
            export_png(self._grid_plot, filename=buff)
            buff.seek(0)
            data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
        w, h = self._grid_plot.plot_width, self._grid_plot.plot_height
        img = data.reshape((int(h), int(w), -1))
        return img

    def _plt_set_figsize(self, width: int, height: int):
        self._grid_plot.width = width
        self._grid_plot.height = height



from __future__ import annotations

import warnings
import weakref
from typing import TYPE_CHECKING

import matplotlib.markers as mmarkers
import matplotlib.pyplot as plt
import matplotlib.transforms as mtransforms
import numpy as np
from matplotlib.collections import PathCollection
from numpy.typing import NDArray
from psygnal import throttled

from whitecanvas.backend.matplotlib._base import MplLayer
from whitecanvas.protocols import MarkersProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle, Symbol
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from matplotlib.backend_bases import MouseEvent as mplMouseEvent

    from whitecanvas.backend.matplotlib.canvas import Canvas


def _get_path(symbol: Symbol):
    marker_obj = mmarkers.MarkerStyle(symbol.value)
    return marker_obj.get_path().transformed(marker_obj.get_transform())


@check_protocol(MarkersProtocol)
class Markers(PathCollection, MplLayer):
    def __init__(self, xdata, ydata):
        offsets = np.stack([xdata, ydata], axis=1)
        self._symbol = Symbol.CIRCLE
        super().__init__(
            (_get_path(self._symbol),),
            sizes=[6] * len(offsets),
            offsets=offsets,
            offset_transform=plt.gca().transData,
        )
        self.set_transform(mtransforms.IdentityTransform())
        self._edge_styles = [LineStyle.SOLID] * len(offsets)
        self._pick_callbacks = []
        self._hover_texts: list[str] | None = None
        self._canvas = lambda: None

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        offsets = self.get_offsets()
        return offsets[:, 0], offsets[:, 1]

    def _plt_set_data(self, xdata, ydata):
        data = np.stack([xdata, ydata], axis=1)
        self.set_offsets(data)

    ##### HasSymbol protocol #####

    def _plt_get_symbol(self) -> Symbol:
        return self._symbol

    def _plt_set_symbol(self, symbol: Symbol):
        path = _get_path(symbol)
        self.set_paths([path] * len(self.get_offsets()))
        self._symbol = symbol

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        return np.sqrt(self.get_sizes())

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(len(self.get_offsets()), size)
        self.set_sizes(size**2)

    ##### HasFaces protocol #####
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        return self.get_facecolor()

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.get_offsets()))
        self.set_facecolor(color)

    def _plt_get_face_hatch(self) -> Hatch:
        return [Hatch(self.get_hatch() or "")] * len(self.get_offsets())

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if not isinstance(pattern, Hatch):
            if len(set(pattern)) > 1:
                warnings.warn(
                    "matplotlib markers do not support multiple hatch patterns.",
                    UserWarning,
                    stacklevel=2,
                )
            pattern = pattern[0]
        if pattern is Hatch.SOLID:
            ptn = None
        else:
            ptn = pattern.value
        self.set_hatch(ptn)

    ##### HasEdges protocol #####

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.get_edgecolor()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.get_offsets()))
        self.set_edgecolor(color)

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return self._edge_styles

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            styles = [style.value] * len(self.get_offsets())
        else:
            styles = [s.value for s in style]
        self.set_linestyle(styles)
        self._edge_styles = styles

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return self.get_linewidth()

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(len(self.get_offsets()), width)
        self.set_linewidth(width)

    def _plt_connect_pick_event(self, callback):
        def cb(event):
            if event.artist is not self:
                return
            callback(event.ind)

        self._pick_callbacks.append(cb)

    def _plt_set_hover_text(self, texts: list[str]):
        self._hover_texts = texts

    def post_add(self, canvas: Canvas):
        fig = self.get_figure()
        for cb in self._pick_callbacks:
            fig.canvas.mpl_connect("pick_event", cb)
        self._canvas = weakref.ref(canvas)
        fig.canvas.mpl_connect(
            "motion_notify_event", throttled(self._on_hover, timeout=500)
        )

    def _on_hover(self, event: mplMouseEvent = None):
        if self._hover_texts is None or not self._plt_get_visible():
            return
        canvas = self._canvas()
        if canvas is None:
            return
        if event.inaxes is not canvas._axes:
            return
        contains, ind = self.contains(event)
        if not contains:
            canvas._hide_tooltip()
            return
        indices = ind["ind"]
        if len(indices) == 0:
            canvas._hide_tooltip()
            return
        index: int = indices[0]
        hover_text = self._hover_texts[index]
        xy = self.get_offsets()[index]
        canvas._set_tooltip(xy, hover_text)



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented
from whitecanvas.protocols import TextProtocol, check_protocol
from whitecanvas.types import Alignment, LineStyle
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(TextProtocol)
class Texts(visuals.Compound):
    def __init__(
        self, x: NDArray[np.floating], y: NDArray[np.floating], text: list[str]
    ):
        super().__init__(
            [SingleText(x0, y0, text0) for x0, y0, text0 in zip(x, y, text)]
        )
        self.unfreeze()

    @property
    def subvisuals(self) -> list[SingleText]:
        return self._subvisuals

    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### TextProtocol #####

    def _plt_get_text(self) -> list[str]:
        return [t.text for t in self.subvisuals]

    def _plt_set_text(self, text: list[str]):
        for t, text0 in zip(self.subvisuals, text):
            t.text = text0

    def _plt_get_ndata(self) -> int:
        return len(self.subvisuals)

    def _plt_get_text_color(self):
        return np.array([t.color for t in self.subvisuals])

    def _plt_set_text_color(self, color):
        color = as_color_array(color, self._plt_get_ndata())
        for t, color0 in zip(self.subvisuals, color):
            t.color = color0

    def _plt_get_text_size(self) -> float:
        return [t.font_size for t in self.subvisuals]

    def _plt_set_text_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(self._plt_get_ndata(), size)
        for t, size0 in zip(self.subvisuals, size):
            t.font_size = size0

    def _plt_get_text_position(
        self,
    ) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
        pos = np.stack([np.array(t.pos[0, 1:]) for t in self.subvisuals], axis=0)
        return pos[:, 0], pos[:, 1]

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.floating], NDArray[np.floating]]
    ):
        for t, x0, y0 in zip(self.subvisuals, *position):
            t.pos = x0, y0

    def _plt_get_text_anchor(self) -> list[Alignment]:
        return [t._alignment for t in self.subvisuals]

    def _plt_set_text_anchor(self, anc: Alignment | list[Alignment]):
        if isinstance(anc, Alignment):
            anc = [anc] * self._plt_get_ndata()
        for t, anc0 in zip(self.subvisuals, anc):
            va, ha = anc0.split()
            t.anchors = va.value, ha.value
            t._alignment = anc0

    def _plt_get_text_rotation(self) -> float:
        return np.array([t.rotation[0] for t in self.subvisuals])

    def _plt_set_text_rotation(self, rotation: float | NDArray[np.floating]):
        if is_real_number(rotation):
            rotation = np.full(self._plt_get_ndata(), rotation)
        for t, rotation0 in zip(self.subvisuals, rotation):
            t.rotation = rotation0

    def _plt_get_text_fontfamily(self) -> list[str]:
        return [t.face for t in self.subvisuals]

    def _plt_set_text_fontfamily(self, fontfamily: str | list[str]):
        if isinstance(fontfamily, str):
            fontfamily = [fontfamily] * self._plt_get_ndata()
        for t, fontfamily0 in zip(self.subvisuals, fontfamily):
            t.face = fontfamily0

    def _plt_get_face_color(self):
        return np.zeros((self._plt_get_ndata(), 4))

    def _plt_set_face_color(self, color):
        pass

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_patterns()

    def _plt_get_edge_color(self):
        return np.zeros((self._plt_get_ndata(), 4))

    def _plt_set_edge_color(self, color):
        pass

    def _plt_get_edge_width(self) -> float:
        return np.zeros(self._plt_get_ndata())

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        pass

    def _plt_get_edge_style(self) -> LineStyle:
        return [LineStyle.SOLID] * self._plt_get_ndata()

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        pass


class SingleText(visuals.Text):
    def __init__(self, x: float, y: float, text: str):
        super().__init__(text=text, anchor_x="left", anchor_y="bottom")
        self._plt_set_text_position([x, y])
        self.unfreeze()
        self._alignment = Alignment.BOTTOM_LEFT

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### TextProtocol #####

    def _plt_get_text(self) -> str:
        return self.text

    def _plt_set_text(self, text: str):
        self.text = text

    def _plt_get_text_color(self):
        return self.color

    def _plt_set_text_color(self, color):
        self.color = color

    def _plt_get_text_size(self) -> float:
        return self.font_size

    def _plt_set_text_size(self, size: float):
        self.font_size = size

    def _plt_get_text_position(self) -> tuple[float, float]:
        return tuple(self.pos[0, 1:])

    def _plt_set_text_position(self, position: tuple[float, float]):
        self.pos = position

    def _plt_get_text_anchor(self) -> Alignment:
        return self._alignment

    def _plt_set_text_anchor(self, anc: Alignment):
        va, ha = anc.split()
        self.anchors = va.value, ha.value
        self._alignment = anc

    def _plt_get_text_rotation(self) -> float:
        return self.rotation[0]

    def _plt_set_text_rotation(self, rotation: float):
        self.rotation = rotation

    def _plt_get_text_fontfamily(self) -> str:
        return self.face

    def _plt_set_text_fontfamily(self, fontfamily: str):
        self.face = fontfamily



from whitecanvas.backend.bokeh.band import Band
from whitecanvas.backend.bokeh.bars import Bars
from whitecanvas.backend.bokeh.canvas import Canvas, CanvasGrid
from whitecanvas.backend.bokeh.image import Image
from whitecanvas.backend.bokeh.line import MonoLine, MultiLine
from whitecanvas.backend.bokeh.markers import Markers
from whitecanvas.backend.bokeh.text import Texts



"""Layer with a dataframe bound to it."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generic, TypeVar

import numpy as np
from cmap import Color

from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers import _mixin
from whitecanvas.layers import group as _lg
from whitecanvas.layers.tabular import _plans as _p
from whitecanvas.layers.tabular import _shared
from whitecanvas.layers.tabular._df_compat import DataFrameWrapper
from whitecanvas.types import (
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas.dataframe._base import CatIterator

    _FE = _mixin.AbstractFaceEdgeMixin[_mixin.FaceNamespace, _mixin.EdgeNamespace]

_DF = TypeVar("_DF")


def _norm_color_hatch(
    color,
    hatch,
    df: DataFrameWrapper[_DF],
) -> tuple[_p.ColorPlan, _p.HatchPlan]:
    color_cov = _shared.ColumnOrValue(color, df)
    if color_cov.is_column:
        color_by = _p.ColorPlan.from_palette(color_cov.columns)
    elif color_cov.value is not None:
        color_by = _p.ColorPlan.from_const(Color(color_cov.value))
    else:
        color_by = _p.ColorPlan.default()
    hatch_cov = _shared.ColumnOrValue(hatch, df)
    if hatch_cov.is_column:
        hatch_by = _p.HatchPlan.new(hatch_cov.columns)
    elif hatch_cov.value is not None:
        hatch_by = _p.HatchPlan.from_const(Hatch(hatch_cov.value))
    else:
        hatch_by = _p.HatchPlan.default()
    return color_by, hatch_by


class _BoxLikeMixin:
    _source: DataFrameWrapper[_DF]

    def __init__(
        self,
        categories: list[tuple],
        splitby: tuple[str, ...],
        color_by: _p.ColorPlan,
        hatch_by: _p.HatchPlan,
    ):
        self._splitby = splitby
        self._categories = categories
        self._color_by = color_by
        self._hatch_by = hatch_by
        self._get_base().face.color = color_by.generate(self._categories, self._splitby)
        self._get_base().face.hatch = hatch_by.generate(self._categories, self._splitby)

    def _get_base(self) -> _FE:
        """Just for typing."""
        return self._base_layer

    def with_color_palette(self, palette) -> Self:
        if self._color_by.is_const():
            raise ValueError("Cannot redraw color for a constant color")
        color_by = _p.ColorPlan.from_palette(self._color_by.by, palette=palette)
        self._get_base().face.color = color_by.generate(self._categories, self._splitby)
        self._color_by = color_by
        return self

    def with_color(self, color: ColorType) -> Self:
        color_by = _p.ColorPlan.from_const(Color(color))
        self._get_base().face.color = color_by.generate(self._categories, self._splitby)
        self._color_by = color_by
        return self

    def with_hatch_palette(self, choices) -> Self:
        if self._hatch_by.is_const():
            raise ValueError("Cannot redraw hatch for a constant hatch")
        hatch_by = _p.HatchPlan.new(self._hatch_by.by, values=choices)
        self._get_base().face.hatch = hatch_by.generate(self._categories, self._splitby)
        self._hatch_by = hatch_by
        return self

    def with_hatch(self, hatch: str | Hatch) -> Self:
        hatch_by = _p.HatchPlan.from_const(Hatch(hatch))
        self._get_base().face.hatch = hatch_by.generate(self._categories, self._splitby)
        self._hatch_by = hatch_by
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
        alpha: float = 1.0,
    ) -> Self:
        self._get_base().with_edge(color=color, width=width, style=style, alpha=alpha)
        return self


class DFViolinPlot(
    _shared.DataFrameLayerWrapper[_lg.ViolinPlot, _DF],
    _BoxLikeMixin,
    Generic[_DF],
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        extent: float = 0.8,
        shape: str = "both",
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _extent = cat.zoom_factor(dodge=dodge) * extent
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.ViolinPlot.from_arrays(
            x, arr, name=name, orient=orient, shape=shape, extent=_extent,
            backend=backend,
        )  # fmt: skip
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)

    @property
    def orient(self) -> Orientation:
        """Orientation of the violins."""
        return self._base_layer.orient

    def with_shift(self, shift: float = 0.0) -> Self:
        for layer in self._base_layer:
            _old = layer.data
            layer.set_data(edge_low=_old.y0 + shift, edge_high=_old.y1 + shift)
        if canvas := self._canvas_ref():
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self


class DFBoxPlot(
    _shared.DataFrameLayerWrapper[_lg.BoxPlot, _DF], _BoxLikeMixin, Generic[_DF]
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        extent: float = 0.8,
        capsize: float = 0.1,
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge,
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _extent = cat.zoom_factor(dodge=dodge) * extent
        _capsize = cat.zoom_factor(dodge=dodge) * capsize
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.BoxPlot.from_arrays(
            x, arr, name=name, orient=orient, capsize=_capsize, extent=_extent,
            backend=backend,
        )  # fmt: skip
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)

    @property
    def orient(self) -> Orientation:
        """Orientation of the violins."""
        return self._base_layer.orient

    def with_shift(
        self,
        shift: float = 0.0,
    ) -> Self:
        self._base_layer.with_shift(shift)
        return self


class _EstimatorMixin(_BoxLikeMixin):
    orient: Orientation

    def est_by_mean(self) -> Self:
        """Set estimator to mean."""

        def est_func(x):
            return np.mean(x)

        return self._update_estimate(est_func)

    def est_by_median(self) -> Self:
        """Set estimator to median."""

        def est_func(x):
            return np.median(x)

        return self._update_estimate(est_func)

    def err_by_sd(self, scale: float = 1.0, *, ddof: int = 1) -> Self:
        """Set error to standard deviation."""

        def err_func(x):
            _mean = np.mean(x)
            _sd = np.std(x, ddof=ddof) * scale
            return _mean - _sd, _mean + _sd

        return self._update_error(err_func)

    def err_by_se(self, scale: float = 1.0, *, ddof: int = 1) -> Self:
        """Set error to standard error."""

        def err_func(x):
            _mean = np.mean(x)
            _er = np.std(x, ddof=ddof) / np.sqrt(len(x)) * scale
            return _mean - _er, _mean + _er

        return self._update_error(err_func)

    def err_by_quantile(self, low: float = 0.25, high: float | None = None) -> Self:
        """Set error to quantile."""
        if low < 0 or low > 1:
            raise ValueError(f"Quantile must be between 0 and 1, got {low}")
        if high is None:
            high = 1 - low
        elif high < 0 or high > 1:
            raise ValueError(f"Quantile must be between 0 and 1, got {high}")

        def err_func(x):
            _qnt = np.quantile(x, [low, high])
            return _qnt[0], _qnt[1]

        return self._update_error(err_func)

    def _update_estimate(self, est_func: Callable[[np.ndarray], float]) -> Self:
        arrays = self._get_arrays()
        est = [est_func(arr) for arr in arrays]
        self._set_estimation_values(est)
        return self

    def _update_error(
        self,
        err_func: Callable[[np.ndarray], tuple[float, float]],
    ) -> Self:
        arrays = self._get_arrays()
        err_low = []
        err_high = []
        for arr in arrays:
            low, high = err_func(arr)
            err_low.append(low)
            err_high.append(high)
        self._set_error_values(err_low, err_high)
        return self


class DFPointPlot(
    _shared.DataFrameLayerWrapper[_lg.LabeledPlot, _DF], _EstimatorMixin, Generic[_DF]
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        capsize: float = 0.1,
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge,
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _capsize = cat.zoom_factor(dodge=dodge) * capsize
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.LabeledPlot.from_arrays(
            x, arr, name=name, orient=orient, capsize=_capsize, backend=backend,
        )  # fmt: skip
        self._arrays = arr
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)
        base.with_edge(color=theme.get_theme().foreground_color)
        self._orient = orient

    @property
    def orient(self) -> Orientation:
        """Orientation of the violins."""
        return self._orient

    def with_shift(
        self,
        shift: float = 0.0,
    ) -> Self:
        base = self._base_layer
        data = base.data
        if self._orient.is_vertical:
            base.set_data(data.x + shift, data.y)
        else:
            base.set_data(data.x, data.y + shift)
        return self

    def _get_arrays(self) -> list[np.ndarray]:
        return self._arrays

    def _set_estimation_values(self, est):
        if self.orient.is_vertical:
            self._base_layer.set_data(ydata=est)
        else:
            self._base_layer.set_data(xdata=est)

    def _set_error_values(self, err_low, err_high):
        mdata = self._base_layer.data
        if self.orient.is_vertical:
            self._base_layer.yerr.set_data(mdata.x, err_low, err_high)
        else:
            self._base_layer.xerr.set_data(err_low, err_high, mdata.y)


class DFBarPlot(
    _shared.DataFrameLayerWrapper[_lg.LabeledBars, _DF], _EstimatorMixin, Generic[_DF]
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        capsize: float = 0.1,
        extent: float = 0.8,
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge,
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _extent = cat.zoom_factor(dodge=dodge) * extent
        _capsize = cat.zoom_factor(dodge=dodge) * capsize
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.LabeledBars.from_arrays(
            x, arr, name=name, orient=orient, capsize=_capsize, extent=_extent,
            backend=backend,
        )  # fmt: skip
        self._arrays = arr
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)
        base.with_edge(color=theme.get_theme().foreground_color)
        self._orient = orient

    @property
    def orient(self) -> Orientation:
        return self._base_layer.bars.orient

    def _get_arrays(self) -> list[np.ndarray]:
        return self._arrays

    def _set_estimation_values(self, est):
        if self.orient.is_vertical:
            self._base_layer.set_data(ydata=est)
        else:
            self._base_layer.set_data(xdata=est)

    def _set_error_values(self, err_low, err_high):
        mdata = self._base_layer.data
        if self.orient.is_vertical:
            self._base_layer.yerr.set_data(mdata.x, err_low, err_high)
        else:
            self._base_layer.xerr.set_data(err_low, err_high, mdata.y)



from __future__ import annotations

from functools import reduce

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from qtpy import QtCore, QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    from_qt_line_style,
    to_qt_line_style,
)
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.types import LineStyle
from whitecanvas.utils.type_check import is_real_number


@check_protocol(LineProtocol)
class MonoLine(pg.PlotCurveItem, PyQtLayer):
    def __init__(self, xdata, ydata):
        pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        pen.setCosmetic(True)
        super().__init__(xdata, ydata, pen=pen, antialias=False)

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self.getData()

    def _plt_set_data(self, xdata, ydata):
        self.setData(xdata, ydata)

    ##### HasEdges #####
    def _get_pen(self) -> QtGui.QPen:
        return self.opts["pen"]

    def _plt_get_edge_width(self) -> float:
        return self._get_pen().widthF()

    def _plt_set_edge_width(self, width: float):
        pen = self._get_pen()
        pen.setWidthF(width)
        self.setPen(pen)

    def _plt_get_edge_style(self) -> LineStyle:
        return from_qt_line_style(self._get_pen().style())

    def _plt_set_edge_style(self, style: LineStyle):
        pen = self._get_pen()
        pen.setStyle(to_qt_line_style(style))
        self.setPen(pen)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(self._get_pen().color().getRgbF())

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        pen = self._get_pen()
        pen.setColor(array_to_qcolor(color))
        self.setPen(pen)

    def _plt_get_antialias(self) -> bool:
        return self.opts["antialias"]

    def _plt_set_antialias(self, antialias: bool):
        self.opts["antialias"] = antialias
        self.update()


@check_protocol(MultiLineProtocol)
class MultiLine(pg.ItemGroup, PyQtLayer):
    def __init__(self, data: list[NDArray[np.floating]]):
        super().__init__()
        pen = QtGui.QPen(QtGui.QColor(255, 255, 255))
        pen.setCosmetic(True)
        self._lines: list[pg.PlotCurveItem] = []
        self._qpens: list[QtGui.QPen] = []
        for seg in data:
            item = pg.PlotCurveItem(seg[:, 0], seg[:, 1], pen=pen, antialias=True)
            self.addItem(item)
            self._lines.append(item)
            self._qpens.append(pen)
        self._data = data
        self._bounding_rect_cache = None
        self._antialias = True

    def boundingRect(self):
        if self._bounding_rect_cache is not None:
            return self._bounding_rect_cache
        if len(self._lines) == 0:
            return QtCore.QRectF()
        rect = reduce(lambda a, b: a | b, (item.boundingRect() for item in self._lines))
        self._bounding_rect_cache = rect
        return rect

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self._data

    def _plt_set_data(self, data: list[NDArray[np.floating]]):
        ndata = len(data)
        nitem = len(self._lines)
        scene = self.scene()
        if ndata < nitem:
            for item in self._lines[ndata:]:
                scene.removeItem(item)
            self._lines = self._lines[:ndata]
            self._qpens = self._qpens[:ndata]
        else:
            # pick last pen
            if nitem > 0:
                pen = self._qpens[-1]
            else:
                pen = QtGui.QPen(QtGui.QColor(255, 255, 255))
            for _ in range(ndata - nitem):
                item = pg.PlotCurveItem(pen=pen, antialias=True)
                self.addItem(item)
                self._lines.append(item)
                self._qpens.append(pen)
        for item, seg in zip(self._lines, data):
            item.setData(seg[:, 0], seg[:, 1])
        self._data = data
        self._bounding_rect_cache = None

    ##### HasEdges #####
    def _set_pen_to_curves(self):
        for item, pen in zip(self._lines, self._qpens):
            item.setPen(pen)

    def _plt_get_edge_width(self) -> NDArray[np.float32]:
        return np.array([pen.widthF() for pen in self._qpens], dtype=np.float32)

    def _plt_set_edge_width(self, width: float | NDArray[np.float32]):
        if is_real_number(width):
            widths: list[float] = [width] * len(self._qpens)
        else:
            widths = width
        for pen, w in zip(self._qpens, widths):
            pen.setWidthF(w)
        self._set_pen_to_curves()

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_qt_line_style(pen.style()) for pen in self._qpens]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            styles: list[LineStyle] = [style] * len(self._qpens)
        else:
            styles = style
        for pen, s in zip(self._qpens, styles):
            pen.setStyle(to_qt_line_style(s))
        self._set_pen_to_curves()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(
            [pen.color().getRgbF() for pen in self._qpens], dtype=np.float32
        )

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            colors = [color] * len(self._qpens)
        else:
            colors = color
        for pen, c in zip(self._qpens, colors):
            pen.setColor(array_to_qcolor(c))
        self._set_pen_to_curves()

    def _plt_get_antialias(self) -> bool:
        return self._antialias

    def _plt_set_antialias(self, antialias: bool):
        for item in self._lines:
            item.opts["antialias"] = antialias
            item.update()
        self._antialias = antialias



from __future__ import annotations

from whitecanvas.types import Hatch, LineStyle


def edge_style():
    def _getter(self):
        return getattr(self, "__edge_style_value", LineStyle.SOLID)

    def _setter(self, value: LineStyle):
        setattr(self, "__edge_style_value", value)

    return _getter, _setter


def edge_styles():
    def _getter(self):
        return getattr(
            self, "__edge_style_value", [LineStyle.SOLID] * self._plt_get_ndata()
        )

    def _setter(self, value: LineStyle | list[LineStyle]):
        if isinstance(value, LineStyle):
            value = [value] * self._plt_get_ndata()
        setattr(self, "__edge_style_value", value)

    return _getter, _setter


def face_pattern():
    def _getter(self):
        return getattr(self, "__face_pattern_value", Hatch.SOLID)

    def _setter(self, value: Hatch):
        setattr(self, "__face_pattern_value", value)

    return _getter, _setter


def face_patterns():
    def _getter(self):
        return getattr(
            self, "__face_pattern_value", [Hatch.SOLID] * self._plt_get_ndata()
        )

    def _setter(self, value: Hatch | list[Hatch]):
        if isinstance(value, Hatch):
            value = [value] * self._plt_get_ndata()
        setattr(self, "__face_pattern_value", value)

    return _getter, _setter



from __future__ import annotations

from typing import Any, overload

import numpy as np
from cmap import Colormap
from numpy.typing import ArrayLike, NDArray
from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers._base import DataBoundLayer, LayerEvents
from whitecanvas.protocols import ImageProtocol
from whitecanvas.types import ArrayLike1D, ColormapType, HistBinType, Origin, _Void
from whitecanvas.utils.normalize import as_array_1d
from whitecanvas.utils.type_check import is_real_number

_void = _Void()


class ImageEvents(LayerEvents):
    cmap = Signal(Colormap)
    clim = Signal(tuple)
    shift = Signal(tuple)
    scale = Signal(tuple)


class Image(DataBoundLayer[ImageProtocol, NDArray[np.number]]):
    """
    Grayscale or RGBA image layer.

    Parameters
    ----------
    image : array_like
        2D or 3D array of image data.
    cmap : colormap type, default "gray"
        Colormap to use.
    clim : tuple of float or None, optional
        Contrast limits. If `None`, the limits are set to the min and max of the data.
        You can also pass `None` separately to either limit to only autoscale one of
        them.
    origin : str or Origin, default "corner"
        Origin of the image. This is a redundant parameter which overlaps with `shift`,
        but it makes it easier to operate on the image.
    """

    events: ImageEvents
    _events_class = ImageEvents

    def __init__(
        self,
        image: ArrayLike,
        *,
        name: str | None = None,
        cmap: ColormapType = "gray",
        clim: tuple[float | None, float | None] | None = None,
        shift: tuple[float, float] = (0, 0),
        scale: tuple[float, float] = (1.0, 1.0),
        backend: Backend | str | None = None,
    ):
        img = _normalize_image(image)
        self._origin = Origin.CORNER
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), img)
        if img.ndim == 3:
            cmap = clim = _void
        self._x_hint = self._y_hint = None
        self.update(cmap=cmap, clim=clim, shift=shift, scale=scale)

    def _get_layer_data(self) -> NDArray[np.number]:
        """Current image data of the layer."""
        return self._backend._plt_get_data()

    def _norm_layer_data(self, data: Any) -> NDArray[np.number]:
        return _normalize_image(data)

    def _set_layer_data(self, data: NDArray[np.number]):
        """Set the data of the layer."""
        self._backend._plt_set_data(data)

    @property
    def cmap(self) -> Colormap:
        """Current colormap."""
        return self._backend._plt_get_colormap()

    @cmap.setter
    def cmap(self, cmap: ColormapType):
        _cmap = Colormap(cmap)
        self._backend._plt_set_colormap(_cmap)
        self.events.cmap.emit(_cmap)

    @property
    def clim(self) -> tuple[float, float]:
        """Current contrast limits."""
        return self._backend._plt_get_clim()

    @clim.setter
    def clim(self, clim: tuple[float | None, float | None] | None):
        if clim is None:
            low, high = None, None
        else:
            low, high = clim
        if low is None:
            low = self.data.min()
        if high is None:
            high = self.data.max()
        self._backend._plt_set_clim((low, high))
        self.events.clim.emit((low, high))

    @property
    def shape(self) -> tuple[int, int]:
        """The visual shape of the image (shape without the color axis)."""
        if self.is_rgba:
            return self.data.shape[:2]
        else:
            return self.data.shape

    @property
    def data_mapped(self) -> NDArray[np.number]:
        """The colored image data (N, M, 4) mapped by the colormap."""
        if self.is_rgba:
            return self.data
        # normalize data to [0, 1]
        cmin, cmax = self.clim
        data_norm = (self.data - cmin) / (cmax - cmin)
        return self.cmap(data_norm)

    @property
    def shift(self) -> tuple[float, float]:
        """Current shift from the origin."""
        shift = self._backend._plt_get_translation()
        sx, sy = self.scale
        if self.origin is Origin.EDGE:
            shift = shift[0] - 0.5 * sx, shift[1] - 0.5 * sy
        elif self.origin is Origin.CORNER:
            pass
        elif self.origin is Origin.CENTER:
            sizex, sizey = self.data.shape[:2]
            shift = shift[0] + (sizex - 1) / 2 * sx, shift[1] + (sizey - 1) / 2 * sy
        else:
            raise RuntimeError("Unreachable")
        return shift

    @shift.setter
    def shift(self, shift: tuple[float, float]):
        img = self.data
        sx, sy = self.scale
        if self.origin is Origin.EDGE:
            shift = shift[0] + 0.5 * sx, shift[1] + 0.5 * sy
        elif self.origin is Origin.CORNER:
            pass
        elif self.origin is Origin.CENTER:
            sizex, sizey = img.shape[:2]
            shift = shift[0] - (sizex - 1) / 2 * sx, shift[1] - (sizey - 1) / 2 * sy
        else:
            raise RuntimeError("Unreachable")
        self._backend._plt_set_translation(shift)
        self._x_hint, self._y_hint = _hint_for(
            (img.shape[1], img.shape[0]), shift=shift, scale=self.scale
        )
        self.events.shift.emit(shift)

    @property
    def shift_raw(self) -> tuple[float, float]:
        """Current shift from the origin as a raw data."""
        return self._backend._plt_get_translation()

    @property
    def scale(self) -> tuple[float, float]:
        """Current scale."""
        return self._backend._plt_get_scale()

    @scale.setter
    def scale(self, scale: float | tuple[float, float]):
        if is_real_number(scale):
            scale = float(scale), float(scale)
        dx, dy = scale
        if dx <= 0 or dy <= 0:
            raise ValueError("Scale must be positive.")
        self._backend._plt_set_scale(scale)
        img = self.data
        self._x_hint, self._y_hint = _hint_for(
            (img.shape[1], img.shape[0]), shift=self.shift, scale=scale
        )
        self.events.scale.emit(scale)

    @property
    def origin(self) -> Origin:
        """Current origin of the image."""
        return self._origin

    @origin.setter
    def origin(self, origin: Origin | str):
        self._origin = Origin(origin)
        self.shift = self.shift  # recalculate

    def update(
        self,
        *,
        cmap: ColormapType | _Void = _void,
        clim: tuple[float | None, float | None] | None | _Void = _void,
        shift: tuple[float, float] | _Void = _void,
        scale: tuple[float, float] | _Void = _void,
        origin: str | Origin | _Void = _void,
    ) -> Image:
        if cmap is not _void:
            if self.is_rgba:
                raise ValueError("Cannot set colormap for an RGBA image.")
            self.cmap = cmap
        if clim is not _void:
            if self.is_rgba:
                raise ValueError("Cannot set contrast limits for an RGBA image.")
            self.clim = clim
        if origin is not _void:
            self.origin = origin
        if shift is not _void:
            self.shift = shift
        if scale is not _void:
            self.scale = scale
        return self

    @overload
    def fit_to(self, bbox: tuple[float, float, float, float], /) -> Image:
        ...

    @overload
    def fit_to(self, x0, y0, x1, y1, /) -> Image:
        ...

    def fit_to(self, *args) -> Image:
        """Fit the image to the given bounding box."""
        if len(args) == 1:
            x0, y0, x1, y1 = args[0]
        elif len(args) == 4:
            x0, y0, x1, y1 = args
        else:
            raise TypeError("fit_to() takes 1 or 4 positional arguments.")
        dx, dy = x1 - x0, y1 - y0
        self.shift = (x0 + 0.5, y0 + 0.5)
        self.scale = (dx / self.data.shape[0], dy / self.data.shape[1])
        return self

    @property
    def is_rgba(self) -> bool:
        """Whether the image is RGBA."""
        return self.data.ndim == 3

    @classmethod
    def build_hist(
        cls,
        x: ArrayLike1D,
        y: ArrayLike1D,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        range=None,
        name: str | None = None,
        density: bool = False,
        cmap: ColormapType = "inferno",
        backend: Backend | str | None = None,
    ):
        _x = as_array_1d(x)
        _y = as_array_1d(y)
        if _x.size != _y.size:
            raise ValueError("x and y must have the same size.")
        if isinstance(bins, (int, np.number, str)):
            xbins = ybins = bins
        else:
            xbins, ybins = bins
        if range is None:
            xrange = yrange = None
        else:
            xrange, yrange = range
        _bins = (
            np.histogram_bin_edges(_x, xbins, xrange),
            np.histogram_bin_edges(_y, ybins, yrange),
        )
        hist, xedges, yedges = np.histogram2d(
            _x, _y, bins=_bins, range=(xrange, yrange), density=density
        )
        hist_t = hist.T
        shift = (xedges[0], yedges[0])
        scale = (xedges[1] - xedges[0], yedges[1] - yedges[0])
        self = cls(
            hist_t, name=name, cmap=cmap, shift=shift, scale=scale, backend=backend
        )
        self.origin = Origin.EDGE
        return self


def _normalize_image(image):
    img = np.asarray(image)
    if img.dtype.kind not in "uif":
        raise TypeError(f"Only numerical arrays are allowed, got {img.dtype}")
    # check shape
    if img.ndim == 2:
        pass
    elif img.ndim == 3:
        nchannels = img.shape[2]
        if nchannels not in (3, 4):
            raise ValueError(
                "If 3D array is given, the last dimension must be 3 or 4, "
                f"got shape {img.shape}."
            )
    else:
        raise ValueError(f"Only 2D or 3D arrays are allowed, got {img.ndim}")
    return img


def _hint_for(
    shape: tuple[int, int],
    shift: tuple[float, float] = (0, 0),
    scale: tuple[float, float] = (1, 1),
) -> tuple[float, float]:
    xhint = np.array([-0.5, shape[0] - 0.5]) * scale[0] + shift[0]
    yhint = np.array([-0.5, shape[1] - 0.5]) * scale[1] + shift[1]
    return tuple(xhint), tuple(yhint)



from __future__ import annotations

from typing import Any, Sequence

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers._primitive import Line
from whitecanvas.layers.group._collections import LayerCollectionBase
from whitecanvas.types import LineStyle, XYData
from whitecanvas.utils.normalize import as_any_1d_array, as_color_array
from whitecanvas.utils.type_check import is_real_number


class LineCollection(LayerCollectionBase[Line]):
    """Collection of lines."""

    def __init__(
        self,
        segments: list[Any],
        name: str | None = None,
        backend: str | Backend | None = None,
    ):
        lines = [Line([], [], backend=backend) for _ in segments]
        for line, seg in zip(lines, segments):
            line.data = seg
        super().__init__(lines, name=name)

    @property
    def data(self) -> list[XYData]:
        return [line.data for line in self]

    @data.setter
    def data(self, data: list[XYData]):
        ndata_in = len(data)
        ndata_now = len(self)
        if ndata_in > ndata_now:
            for _ in range(ndata_now, ndata_in):
                self.append(Line([], []))
        elif ndata_in < ndata_now:
            for _ in range(ndata_in, ndata_now):
                del self[-1]
        for line, d in zip(self, data):
            line.data = d

    @property
    def width(self) -> NDArray[np.float32]:
        return np.array([line.width for line in self], dtype=np.float32)

    @width.setter
    def width(self, width: float | Sequence[float]):
        if is_real_number(width):
            _width = [width] * len(self)
        else:
            _width = np.asarray(width, dtype=np.float32)
        if len(_width) != len(self):
            raise ValueError(
                f"width must be a float or a sequence of length {len(self)}"
            )
        for line, w in zip(self, _width):
            line.width = w

    @property
    def color(self) -> NDArray[np.float32]:
        return np.array([line.color for line in self], dtype=np.float32)

    @color.setter
    def color(self, color: str | Sequence[str]):
        col = as_color_array(color, len(self))
        for line, c in zip(self, col):
            line.color = c

    @property
    def style(self) -> list[LineStyle]:
        return np.array([line.style for line in self], dtype=np.float32)

    @style.setter
    def style(self, style: str | Sequence[str]):
        styles = as_any_1d_array(style, len(self))
        for line, s in zip(self, styles):
            line.style = s



"""
Plot API for whitecanvas.

>>> import numpy as np
>>> from whitecanvas import plot as plt
>>> plt.line(np.sin([1, 2, 3])).with_markers()
>>> plt.show()
"""

from whitecanvas.core import new_canvas
from whitecanvas.plot._canvases import current_canvas, current_grid, show, subplots
from whitecanvas.plot._methods import (
    band,
    bars,
    cat,
    errorbars,
    hist,
    infcurve,
    infline,
    kde,
    line,
    markers,
    rug,
    spans,
    text,
)

figure = new_canvas

__all__ = [
    "line",
    "markers",
    "bars",
    "band",
    "hist",
    "spans",
    "infcurve",
    "infline",
    "errorbars",
    "kde",
    "rug",
    "text",
    "cat",
    "figure",
    "show",
    "subplots",
    "current_grid",
    "current_canvas",
]



import numpy as np
import pytest
from numpy.testing import assert_allclose

from whitecanvas.layers.tabular import _plans as _p


@pytest.mark.parametrize(
    "column, expected, policy",
    [
        ([1, 3, 2], [0, 1, 2], _p.NoMarginPolicy()),
        (["T", "TT", "Q"], [0, 1, 2], _p.NoMarginPolicy()),
        ([1, 3, 2], [0, 1.3, 2.6], _p.ConstMarginPolicy(0.3)),
    ]
)
def test_offset_one_level(column, expected, policy):
    plan = _p.OffsetPlan(("a",), [policy])
    out = plan.generate(np.array(column).reshape(-1, 1), by_all=("a",))
    assert_allclose(out, expected)

@pytest.mark.parametrize(
    "arr, expected, policy",
    [
        ([[1, 1], [1, 2], [1, 3], [2, 1]], [0, 0, 0, 1], _p.NoMarginPolicy()),
        ([[1, 1], [1, 2], [1, 3], [2, 3]], [0, 0, 0, 1], _p.NoMarginPolicy()),
        ([[1, 1], [1, 2], [1, 3], [2, 3]], [0, 0, 0, 1.3], _p.ConstMarginPolicy(0.3)),
    ]
)
def test_offset_subset(arr, expected, policy):
    plan = _p.OffsetPlan(("a",), [policy])
    out = plan.generate(np.array(arr), by_all=("a", "b"))
    assert_allclose(out, expected)

@pytest.mark.parametrize(
    "arr, expected, policy",
    [
        ([[1, 1], [1, 2], [1, 3], [2, 1]], [0, 1, 2, 3],
         [_p.NoMarginPolicy(), _p.NoMarginPolicy()]),
        ([[1, 1], [1, 2], [1, 3], [2, 3]], [0, 1, 2, 5],
         [_p.NoMarginPolicy(), _p.NoMarginPolicy()]),
        ([[1, 1], [1, 2], [2, 1], [2, 2]], [0, 1, 2.3, 3.3],
         [_p.ConstMarginPolicy(0.3), _p.NoMarginPolicy()]),
    ]
)  # fmt: skip
def test_offset_multilevel(arr, expected, policy):
    plan = _p.OffsetPlan(("a", "b"), policy)
    out = plan.generate(np.array(arr), by_all=("a", "b"))
    assert_allclose(out, expected)



from __future__ import annotations


class PyQtLayer:
    def _plt_get_visible(self) -> bool:
        return self.isVisible()

    def _plt_set_visible(self, visible: bool):
        self.setVisible(visible)

    def _plt_set_zorder(self, zorder: int):
        self.setZValue(zorder)



from whitecanvas.canvas._base import Canvas, CanvasBase
from whitecanvas.canvas._grid import CanvasGrid, CanvasHGrid, CanvasVGrid, SingleCanvas

__all__ = [
    "CanvasBase",
    "Canvas",
    "CanvasGrid",
    "CanvasHGrid",
    "CanvasVGrid",
    "SingleCanvas",
]



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented
from whitecanvas.protocols import BandProtocol, check_protocol
from whitecanvas.types import LineStyle, Orientation


@check_protocol(BandProtocol)
class Band(visuals.Polygon):
    def __init__(self, t, ydata0, ydata1, orient: Orientation):
        if orient.is_vertical:
            fw = np.stack([t, ydata0], axis=1)
            bw = np.stack([t[::-1], ydata1[::-1]], axis=1)
        else:
            fw = np.stack([ydata0, t], axis=1)
            bw = np.stack([ydata1[::-1], t[::-1]], axis=1)
        verts = np.concatenate([fw, bw], axis=0)
        self._edge_style = LineStyle.SOLID
        try:
            super().__init__(verts, border_width=0)
        except Exception:
            super().__init__(verts, border_width=0, triangulate=False)
        self.unfreeze()
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    ##### LayerProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_vertical_data(self):
        return self._t, self._y0, self._y1

    def _plt_get_horizontal_data(self):
        return self._t, self._y0, self._y1

    def _plt_set_vertical_data(self, t, ydata0, ydata1):
        verts = np.concatenate(
            [
                np.stack([t, ydata0], axis=1),
                np.stack([t[::-1], ydata1[::-1]], axis=1),
            ],
            axis=0,
        )
        self.pos = verts
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    def _plt_set_horizontal_data(self, t, ydata0, ydata1):
        verts = np.concatenate(
            [
                np.stack([ydata0, t], axis=1),
                np.stack([ydata1[::-1], t[::-1]], axis=1),
            ],
            axis=0,
        )
        self.pos = verts
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    ##### HasFace protocol #####
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        return np.array(self.color, dtype=np.float32)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        self.color = color

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_pattern()

    ##### HasEdges protocol #####
    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(self.border_color, dtype=np.float32)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.border_color = color

    def _plt_get_edge_width(self) -> float:
        return self.border.width

    def _plt_set_edge_width(self, width: float):
        self.border.set_data(width=width)

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_style()



from whitecanvas.backend.pyqtgraph.band import Band
from whitecanvas.backend.pyqtgraph.bars import Bars
from whitecanvas.backend.pyqtgraph.canvas import Canvas, CanvasGrid
from whitecanvas.backend.pyqtgraph.image import Image
from whitecanvas.backend.pyqtgraph.line import MonoLine, MultiLine
from whitecanvas.backend.pyqtgraph.markers import Markers
from whitecanvas.backend.pyqtgraph.text import Texts



from __future__ import annotations


class MplLayer:
    def _plt_get_visible(self):
        return self.get_visible()

    def _plt_set_visible(self, visible):
        self.set_visible(visible)

    def _plt_set_zorder(self, zorder: int):
        self.set_zorder(zorder)


OVERLAY_ZORDER = 10000


def as_overlay(layer: MplLayer, canvas):
    layer.set_transform(canvas._axes.transAxes)
    layer.set_zorder(OVERLAY_ZORDER)



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.utils.normalize import as_array_1d, as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(LineProtocol)
class MonoLine(visuals.Line):
    def __init__(self, xdata, ydata):
        data = np.stack([xdata, ydata], axis=1)
        super().__init__(data, antialias=True)
        self.unfreeze()

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        pos = self.pos
        return pos[:, 0], pos[:, 1]

    def _plt_set_data(self, xdata, ydata):
        self.set_data(np.stack([xdata, ydata], axis=1))

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> float:
        return self.width

    def _plt_set_edge_width(self, width: float):
        self.set_data(width=width)

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_style()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.color

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.set_data(color=color)

    def _plt_get_antialias(self) -> bool:
        return self.antialias

    def _plt_set_antialias(self, antialias: bool):
        self.antialias = antialias


@check_protocol(MultiLineProtocol)
class MultiLine(visuals.Compound):
    def __init__(self, data: list[NDArray[np.float32]]):
        items = []
        for seg in data:
            item = visuals.Line(seg, width=1, antialias=True)
            items.append(item)
        self._data = data
        self._antialias = True
        super().__init__(items)
        self.unfreeze()

    @property
    def _lines(self) -> list[visuals.Line]:
        return self._subvisuals

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self._data

    def _plt_set_data(self, data: list[NDArray[np.floating]]):
        ndata = len(data)
        nitem = len(self._lines)
        if ndata < nitem:
            for item in self._lines[ndata:]:
                self.remove_subvisual(item)
        else:
            for _ in range(ndata - nitem):
                item = visuals.Line()
                self.add_subvisual(item)
        for item, seg in zip(self._lines, data):
            item.set_data(seg)
        self._data = data

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        if len(self._lines) == 0:
            return np.zeros(0, dtype=np.float32)
        return np.array([line.width for line in self._lines], dtype=np.float32)

    def _plt_set_edge_width(self, width):
        if is_real_number(width):
            for item in self._lines:
                item.set_data(width=width)
        else:
            widths = as_array_1d(width)
            if widths.size != len(self._lines):
                raise ValueError(
                    f"width must be a scalar or an array of length {len(self._lines)}"
                )
            for item, w in zip(self._lines, widths):
                item.set_data(width=w)

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_style()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        if len(self._lines) == 0:
            return np.zeros((0, 4), dtype=np.float32)
        return np.stack([line.color for line in self._lines], axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        col = as_color_array(color, len(self._lines))
        for item, c in zip(self._lines, col):
            item.set_data(color=c)

    def _plt_get_antialias(self) -> bool:
        return self._antialias

    def _plt_set_antialias(self, antialias: bool):
        for item in self._lines:
            item.antialias = antialias
        self._antialias = antialias



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

from matplotlib import pyplot as plt

from whitecanvas.types import LineStyle

if TYPE_CHECKING:
    from whitecanvas.backend.matplotlib.canvas import Canvas


class SupportsText:
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)
        self._text = ""
        self._fontdict = {
            "color": plt.rcParams["text.color"],
            "fontsize": plt.rcParams["font.size"],
            "family": plt.rcParams["font.family"],
        }

    def _plt_get_color(self):
        return self._fontdict["color"]

    def _plt_get_size(self) -> str:
        return self._fontdict["fontsize"]

    def _plt_get_fontfamily(self) -> str:
        return self._fontdict["family"]


class Title(SupportsText):
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)
        self._text = ""
        self._fontdict = {
            "color": plt.rcParams["text.color"],
            "fontsize": plt.rcParams["font.size"],
            "family": plt.rcParams["font.family"],
        }

    def _plt_get_text(self) -> str:
        return self._canvas()._axes.get_title()

    def _plt_set_text(self, text: str):
        self._canvas()._axes.set_title(text, self._fontdict)
        self._text = text

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        self._canvas()._axes.set_title(self._text, fontdict=d)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        return bool(self._canvas()._axes.get_title())

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._canvas()._axes.set_title(self._text, fontdict=self._fontdict)
        else:
            self._canvas()._axes.set_title("")

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        self._canvas()._axes.set_title(self._text, fontdict=d)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        self._canvas()._axes.set_title(self._text, fontdict=d)
        self._fontdict = d


class XLabel(SupportsText):
    def _plt_get_text(self) -> str:
        return self._canvas()._axes.get_xlabel()

    def _plt_set_text(self, text: str):
        self._canvas()._axes.set_xlabel(text, self._fontdict)
        self._text = text

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        self._canvas()._axes.set_xlabel(self._text, fontdict=d)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        return bool(self._canvas()._axes.get_xlabel())

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._canvas()._axes.set_xlabel(self._text, fontdict=self._fontdict)
        else:
            self._canvas()._axes.set_xlabel("")

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        self._canvas()._axes.set_xlabel(self._text, fontdict=d)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        self._canvas()._axes.set_xlabel(self._text, fontdict=d)
        self._fontdict = d


class YLabel(SupportsText):
    def _plt_get_text(self) -> str:
        return self._canvas()._axes.get_ylabel()

    def _plt_set_text(self, text: str):
        self._canvas()._axes.set_ylabel(text, self._fontdict)
        self._text = text

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        self._canvas()._axes.set_ylabel(self._text, fontdict=d)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        return bool(self._canvas()._axes.get_ylabel())

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._canvas()._axes.set_ylabel(self._text, fontdict=self._fontdict)
        else:
            self._canvas()._axes.set_ylabel("")

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        self._canvas()._axes.set_ylabel(self._text, fontdict=d)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        self._canvas()._axes.set_ylabel(self._text, fontdict=d)
        self._fontdict = d


class XTicks(SupportsText):
    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        axes = self._canvas()._axes
        return axes.get_xticks(), [x.get_text() for x in axes.get_xticklabels()]

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        self._canvas()._axes.set_xticks(pos, labels)

    def _plt_reset_override(self):
        self._canvas()._axes.set_xticks([])

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        for x in self._canvas()._axes.get_xticklabels():
            x.set_color(color)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        return self._canvas()._axes.get_xticklines()[0].get_visible()

    def _plt_set_visible(self, visible: bool):
        axes = self._canvas()._axes
        for tick in axes.get_xticklines():
            tick.set_visible(visible)
        for text in axes.get_xticklabels():
            text.set_visible(visible)

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        for x in self._canvas()._axes.get_xticklabels():
            x.set_fontsize(size)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        for x in self._canvas()._axes.get_xticklabels():
            x.set_fontfamily(font)
        self._fontdict = d

    def _plt_get_text_rotation(self) -> float:
        return self._canvas()._axes.get_xticklabels()[0].get_rotation()

    def _plt_set_text_rotation(self, rotation: float):
        for x in self._canvas()._axes.get_xticklabels():
            x.set_rotation(rotation)


class YTicks(SupportsText):
    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        axes = self._canvas()._axes
        return axes.get_yticks(), [x.get_text() for x in axes.get_yticklabels()]

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        self._canvas()._axes.set_yticks(pos, labels)

    def _plt_reset_override(self):
        self._canvas()._axes.set_yticks([])

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        for x in self._canvas()._axes.get_yticklabels():
            x.set_color(color)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        return self._canvas()._axes.get_yticklines()[0].get_visible()

    def _plt_set_visible(self, visible: bool):
        axes = self._canvas()._axes
        for tick in axes.get_yticklines():
            tick.set_visible(visible)
        for text in axes.get_yticklabels():
            text.set_visible(visible)

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        for x in self._canvas()._axes.get_yticklabels():
            x.set_fontsize(size)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        for x in self._canvas()._axes.get_yticklabels():
            x.set_fontfamily(font)
        self._fontdict = d


class AxisBase:
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)


class XAxis(AxisBase):
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)

    def _plt_get_limits(self) -> tuple[float, float]:
        axes = self._canvas()._axes
        x0, x1 = axes.get_xlim()
        if axes.xaxis_inverted():
            return x1, x0
        else:
            return x0, x1

    def _plt_set_limits(self, limits: tuple[float, float]):
        axes = self._canvas()._axes
        if axes.xaxis_inverted():
            limits = limits[::-1]
        axes.set_xlim(*limits)

    def _plt_get_color(self):
        return self._canvas()._axes.xaxis.get_tick_params()["color"]

    def _plt_set_color(self, color):
        ax = self._canvas()._axes
        color = tuple(color)
        ax.xaxis.set_tick_params(color=color, labelcolor=color)
        ax.spines["bottom"].set_color(color)

    def _plt_flip(self):
        self._canvas()._axes.invert_xaxis()

    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        self._canvas()._axes.xaxis.grid(
            visible,
            which="major",
            color=color,
            linestyle=style.value,
            linewidth=width,
        )


class YAxis(AxisBase):
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)

    def _plt_get_limits(self) -> tuple[float, float]:
        axes = self._canvas()._axes
        y0, y1 = axes.get_ylim()
        if axes.yaxis_inverted():
            return y1, y0
        else:
            return y0, y1

    def _plt_set_limits(self, limits: tuple[float, float]):
        axes = self._canvas()._axes
        if axes.yaxis_inverted():
            limits = limits[::-1]
        axes.set_ylim(*limits)

    def _plt_get_color(self):
        return self._canvas()._axes.yaxis.get_tick_params()["color"]

    def _plt_set_color(self, color):
        ax = self._canvas()._axes
        color = tuple(color)
        ax.yaxis.set_tick_params(color=color, labelcolor=color)
        ax.spines["left"].set_color(color)

    def _plt_flip(self):
        self._canvas()._axes.invert_yaxis()

    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        self._canvas()._axes.yaxis.grid(
            visible,
            which="major",
            color=color,
            linestyle=style.value,
            linewidth=width,
        )



from __future__ import annotations

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from qtpy import QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    from_qt_brush_style,
    from_qt_line_style,
    to_qt_brush_style,
    to_qt_line_style,
)
from whitecanvas.protocols import BarProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle
from whitecanvas.utils.normalize import as_color_array


@check_protocol(BarProtocol)
class Bars(pg.BarGraphItem, PyQtLayer):
    def __init__(self, xlow, xhigh, ylow, yhigh):
        pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        pen.setCosmetic(True)
        ndata = len(xlow)
        super().__init__(
            x0=xlow, x1=xhigh, y0=ylow, y1=yhigh,
            pens=[QtGui.QPen(pen) for _ in range(ndata)],
            brushes=[QtGui.QBrush(QtGui.QColor(0, 0, 0)) for _ in range(ndata)],
        )  # fmt: skip

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self.opts["x0"], self.opts["x1"], self.opts["y0"], self.opts["y1"]

    def _plt_set_data(self, xlow, xhigh, ylow, yhigh):
        self.setOpts(x0=xlow, x1=xhigh, y0=ylow, y1=yhigh)

    ##### HasFace protocol #####

    def _get_brushes(self) -> list[QtGui.QBrush]:
        return self.opts["brushes"]

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        colors = []
        for brush in self._get_brushes():
            colors.append(brush.color().getRgbF())
        return np.array(colors, dtype=np.float32)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.opts["x0"]))
        brushes = self._get_brushes()
        for brush, c in zip(brushes, color):
            brush.setColor(array_to_qcolor(c))
        self.setOpts(brushes=brushes)

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [from_qt_brush_style(brush.style()) for brush in self._get_brushes()]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        brushes = self._get_brushes()
        if isinstance(pattern, Hatch):
            ptn = to_qt_brush_style(pattern)
            for brush in brushes:
                brush.setStyle(ptn)
        else:
            for brush, ptn in zip(brushes, pattern):
                brush.setStyle(to_qt_brush_style(ptn))
        self.setOpts(brushes=brushes)

    ##### HasEdges protocol #####

    def _get_pens(self) -> list[QtGui.QPen]:
        return self.opts["pens"]

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        colors = []
        for pen in self._get_pens():
            colors.append(pen.color().getRgbF())
        return np.array(colors, dtype=np.float32)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.opts["x0"]))
        pens = self._get_pens()
        for pen, c in zip(pens, color):
            pen.setColor(array_to_qcolor(c))
        self.setOpts(pens=pens)

    def _plt_get_edge_width(self) -> float:
        return np.array([pen.widthF() for pen in self._get_pens()])

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if np.isscalar(width):
            width = np.full(len(self.opts["x0"]), width)
        pens = self._get_pens()
        for pen, w in zip(pens, width):
            pen.setWidthF(w)
        self.setOpts(pens=pens)

    def _plt_get_edge_style(self) -> LineStyle:
        return [from_qt_line_style(pen.style()) for pen in self._get_pens()]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        pens = self._get_pens()
        if isinstance(style, LineStyle):
            style = to_qt_line_style(style)
            for pen in pens:
                pen.setStyle(style)
        else:
            for pen, s in zip(pens, style):
                pen.setStyle(to_qt_line_style(s))
        self.setOpts(pens=pens)



from __future__ import annotations

import bokeh.models as bk_models
import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend.bokeh._base import (
    BokehLayer,
    from_bokeh_hatch,
    from_bokeh_line_style,
    to_bokeh_hatch,
    to_bokeh_line_style,
)
from whitecanvas.protocols import TextProtocol, check_protocol
from whitecanvas.types import Alignment, Hatch, LineStyle
from whitecanvas.utils.normalize import arr_color, as_color_array, hex_color
from whitecanvas.utils.type_check import is_real_number


@check_protocol(TextProtocol)
class Texts(BokehLayer[bk_models.Text]):
    def __init__(
        self, x: NDArray[np.floating], y: NDArray[np.floating], text: list[str]
    ):
        ntexts = len(text)
        self._data = bk_models.ColumnDataSource(
            data={
                "x": x,
                "y": y,
                "text": text,
                "text_font": ["Arial"] * ntexts,
                "text_font_size": ["12pt"] * ntexts,
                "text_color": ["black"] * ntexts,
                "angle": [0] * ntexts,
                "background_fill_color": ["#00000000"] * ntexts,
                "background_hatch_pattern": [""] * ntexts,
                "border_line_color": ["#00000000"] * ntexts,
                "border_line_width": [0] * ntexts,
                "border_line_dash": ["solid"] * ntexts,
            }
        )
        self._model = bk_models.Text(
            x="x",
            y="y",
            text="text",
            text_font="text_font",
            text_font_size="text_font_size",
            text_color="text_color",
            angle="angle",
            background_fill_color="background_fill_color",
            background_hatch_pattern="background_hatch_pattern",
            border_line_color="border_line_color",
            border_line_width="border_line_width",
            border_line_dash="border_line_dash",
        )
        self._visible = True

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._visible = visible
        if visible:
            self._model.text_color = "#00000000"
            self._model.background_fill_color = "#00000000"
            self._model.border_line_color = "#00000000"
        else:
            self._model.text_color = "text_color"
            self._model.background_fill_color = "background_fill_color"
            self._model.border_line_color = "border_line_color"

    ##### TextProtocol #####

    def _plt_get_text(self) -> list[str]:
        return self._data.data["text"]

    def _plt_set_text(self, text: list[str]):
        self._data.data["text"] = text

    def _plt_get_text_color(self):
        return np.stack([arr_color(c) for c in self._data.data["text_color"]])

    def _plt_set_text_color(self, color):
        color = as_color_array(color, len(self._data.data["text"]))
        self._data.data["text_color"] = [hex_color(c) for c in color]

    def _plt_get_text_size(self) -> float:
        return np.array(
            [int(s.rstrip("pt")) for s in self._data.data["text_font_size"]],
            dtype=np.float32,
        )

    def _plt_set_text_size(self, size: float):
        if is_real_number(size):
            size = np.full(len(self._data.data["text"]), size)
        self._data.data["text_font_size"] = [f"{round(s, 1)}pt" for s in size]

    def _plt_get_text_position(
        self,
    ) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
        return self._model.x, self._model.y

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.floating], NDArray[np.floating]]
    ):
        self._data.data["x"], self._data.data["y"] = position

    def _plt_get_text_anchor(self) -> list[Alignment]:
        return [Alignment(anc) for anc in self._data.data["text_align"]]

    def _plt_set_text_anchor(self, anc: Alignment | list[Alignment]):
        if isinstance(anc, Alignment):
            anc = [anc] * len(self._data.data["text"])
        self._data.data["text_align"] = [a.value for a in anc]

    def _plt_get_text_rotation(self) -> NDArray[np.floating]:
        return np.array(self._data.data["angle"], dtype=np.float32)

    def _plt_set_text_rotation(self, rotation: float | NDArray[np.floating]):
        if is_real_number(rotation):
            rotation = np.full(len(self._data.data["text"]), rotation)
        self._data.data["angle"] = rotation

    def _plt_get_text_fontfamily(self) -> list[str]:
        return self._data.data["text_font"]

    def _plt_set_text_fontfamily(self, fontfamily: str | list[str]):
        if isinstance(fontfamily, str):
            fontfamily = [fontfamily] * len(self._data.data["text"])
        self._data.data["text_font"] = fontfamily

    ##### HasFaces #####

    def _plt_get_face_color(self):
        return np.stack(
            [arr_color(c) for c in self._data.data["background_fill_color"]]
        )

    def _plt_set_face_color(self, color):
        color = as_color_array(color, len(self._data.data["text"]))
        self._data.data["background_fill_color"] = [hex_color(c) for c in color]

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [
            from_bokeh_hatch(h) for h in self._data.data["background_hatch_pattern"]
        ]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if isinstance(pattern, Hatch):
            pattern = [pattern] * len(self._data.data["text"])
        self._data.data["background_hatch_pattern"] = [
            to_bokeh_hatch(p) for p in pattern
        ]

    def _plt_get_edge_color(self):
        return np.stack([arr_color(c) for c in self._data.data["border_line_color"]])

    def _plt_set_edge_color(self, color):
        self._model.border_line_color = hex_color(color)

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return np.array(self._data.data["border_line_width"], dtype=np.float32)

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(len(self._data.data["text"]), width)
        self._data.data["border_line_width"] = width

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_bokeh_line_style(s) for s in self._data.data["border_line_dash"]]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            style = [style] * len(self._data.data["text"])
        self._data.data["border_line_dash"] = [to_bokeh_line_style(s) for s in style]



from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, Iterator, Sequence, TypeVar

import numpy as np
from cmap import Color, Colormap
from numpy.typing import NDArray

from whitecanvas.canvas._palette import ColorPalette
from whitecanvas.layers.tabular._utils import unique
from whitecanvas.types import Hatch, LineStyle, Symbol

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers.tabular._df_compat import DataFrameWrapper

_V = TypeVar("_V", bound=Any)


class Plan(ABC, Generic[_V]):
    @abstractmethod
    def generate(
        self,
        vals: np.ndarray,
        by_all: tuple[str, ...],
    ) -> list[_V]:
        """
        Generate values for given labels.

        Length of the returned list must be equal to vals.shape[0].
        """


class CategoricalPlan(Plan[_V]):
    def __init__(self, by: tuple[str, ...]):
        self._by = by

    @property
    def by(self) -> tuple[str, ...]:
        return self._by

    @classmethod
    @abstractmethod
    def default(cls, by: Sequence[str]) -> Self:
        """Create a default plan."""


class OffsetPolicy(ABC):
    """Class that defines how to define offsets for given data."""

    @abstractmethod
    def get(self, interval: int) -> float:
        """Get increment of position for given interval from the previous position."""

    def with_shift(self, val: float) -> CompositeOffsetPolicy:
        return CompositeOffsetPolicy([self, ConstOffset(val)])


class ConstMarginPolicy(OffsetPolicy):
    def __init__(self, margin: float) -> None:
        self._margin = margin

    def get(self, interval: int) -> float:
        return interval + self._margin


class ConstPolicy(OffsetPolicy):
    def __init__(self, incr: float) -> None:
        self._incr = incr

    def get(self, interval: int) -> float:
        return self._incr


class NoMarginPolicy(ConstMarginPolicy):
    def __init__(self) -> None:
        super().__init__(0.0)


class OverlayPolicy(OffsetPolicy):
    def get(self, interval: int) -> float:
        return 0.0


class CompositeOffsetPolicy(OffsetPolicy):
    def __init__(self, policies: list[OffsetPolicy]) -> None:
        self._policies = policies

    def get(self, interval: int) -> float:
        return sum(policy.get(interval) for policy in self._policies)

    def with_shift(self, val: float) -> CompositeOffsetPolicy:
        return CompositeOffsetPolicy([*self._policies, ConstOffset(val)])


class ConstOffset(OffsetPolicy):
    def __init__(self, val: float) -> None:
        self._val = val

    def get(self, idx: int, size: int):
        return self._val


class NoOffset(ConstOffset):
    def __init__(self) -> None:
        super().__init__(0.0)


class ManualOffset(OffsetPolicy):
    def __init__(self, offsets: list[float]) -> None:
        self._offsets = offsets

    def get(self, idx: int, size: int):
        return self._offsets[idx]


class OffsetPlan(CategoricalPlan[float]):
    """Plan of how to define x-offsets for each category."""

    def __init__(self, by: tuple[str, ...], offsets: list[OffsetPolicy]):
        super().__init__(by)
        self._offsets = offsets

    @classmethod
    def default(cls) -> OffsetPlan:
        return cls((), [])

    def more_by(self, *by: str, margin: float = 0.0) -> OffsetPlan:
        """Add more offsets."""
        found = [b for b in by if b in self._by]
        if found:
            raise ValueError(f"{found} is already in the plan.")
        new = [ConstMarginPolicy(margin) for _ in by]
        return OffsetPlan(self._by + tuple(by), self._offsets + new)

    def shift(self, val: float) -> OffsetPlan:
        new = [offset.with_shift(val) for offset in self._offsets]
        return OffsetPlan(self._by, new)

    def iter_ticks(
        self,
        labels: list[tuple[Any, ...]],
        by_all: tuple[str, ...],
    ) -> Iterator[tuple[float, list[str]]]:
        indices = [by_all.index(b) for b in self.by]
        ncols = len(indices)
        out_lookup: dict[tuple, float] = {}
        # make a full mesh where all combinations are included
        each_uniques = [
            unique(np.array([row[i] for row in labels]), axis=None) for i in indices
        ]

        last_row: tuple | None = None
        last_x = np.zeros(ncols, dtype=np.float32)
        # intervals will be like:
        # each_uniques --> intervals
        # [[1, 2], [4, 5, 6]] --> [3, 1]
        # [[1, 2], [4, 5, 6], [7, 8]] --> [6, 2, 1]
        intervals = _to_intervals(each_uniques)
        for row in itertools.product(*each_uniques):
            row_arr = np.array(row)
            if last_row is None:
                last_row = row_arr
                x = 0.0
            else:
                i = int(np.where(row_arr != last_row)[0][0])
                x0 = last_x[i]
                x = x0 + self._offsets[i].get(intervals[i])
                last_x[i:] = x
            out_lookup[row] = x
            last_row = row_arr
        # yield like this:
        # 0.0 ['Female', 'Dinner', 'No']
        # 1.0 ['Female', 'Dinner', 'Yes']
        # 2.0 ['Female', 'Lunch', 'No']
        # 3.0 ['Female', 'Lunch', 'Yes']
        # 4.0 ['Male', 'Dinner', 'No']
        # 5.0 ['Male', 'Dinner', 'Yes']
        # 6.0 ['Male', 'Lunch', 'No']
        # 7.0 ['Male', 'Lunch', 'Yes']
        for _r in labels:
            _pos = out_lookup[tuple(_r[i] for i in indices)]
            _labels = [str(_r[i]) for i in indices]
            yield _pos, _labels

    def generate(
        self,
        labels: list[tuple[Any, ...]],
        by_all: tuple[str, ...],
    ) -> list[float]:
        return [pos for pos, _ in self.iter_ticks(labels, by_all)]


# class TightOffsetPlan(OffsetPlan):

#     def generate(
#         self,
#         labels: np.ndarray,
#         by_all: tuple[str, ...],
#     ) -> list[float]:
#         indices = [by_all.index(b) for b in self.by]
#         ncols = len(indices)
#         out_lookup = {}
#         filt = _filter_unique(labels, indices)
#         x = 0.0
#         last_row = None
#         last_idx = np.zeros(ncols, dtype=np.int32)
#         for idx, row in enumerate(filt):
#             if last_row is None:
#                 last_row = row
#             else:
#                 i: int = np.where(row != last_row)[0][0]
#                 interval = idx - last_idx[i]
#                 offset = self._offsets[i].get(interval)
#                 last_idx[i:] = idx
#                 x += offset
#             out_lookup[tuple(row)] = x
#         out = [out_lookup[tuple(row_all[indices])] for row_all in labels]
#         return out


class CyclicPlan(CategoricalPlan[_V]):
    """
    Plan of cyclic values.

    This object is used to generate values for each category.
    """

    def __init__(self, by: tuple[str, ...], values: list[_V]):
        super().__init__(by)
        self._values = values

    def __repr__(self) -> str:
        cname = type(self).__name__
        return f"{cname}(by={self._by!r}, values={self._values})"

    @property
    def values(self) -> list[_V]:
        return self._values

    @classmethod
    @abstractmethod
    def _default_values(cls) -> list[_V]:
        """Default values for the plan."""

    @classmethod
    @abstractmethod
    def _norm_value(cls, v: Any) -> _V:
        """Normalize the value."""

    @classmethod
    def default(cls) -> Self:
        return cls((), cls._default_values())

    @classmethod
    def new(cls, by, values: list[_V] | None = None):
        """Create a new plan."""
        if values is None:
            values = cls._default_values()
        return cls(tuple(by), values)

    @classmethod
    def from_const(cls, value: _V) -> Self:
        """Create a plan that always returns a constant value."""
        return cls((), values=[cls._norm_value(value)])

    def is_const(self) -> bool:
        """Return True if the plan is a constant plan."""
        return len(self.by) == 0

    def generate(
        self,
        labels: list[tuple[Any, ...]],
        by_all: tuple[str, ...],
    ) -> list[_V]:
        # labels = [("a",), ("b",)] ... the all set of unique labels
        # by_all = ("column-0", "column-1")
        indices = [by_all.index(b) for b in self.by]
        size = len(self.values)
        # filt = _filter_unique(labels, indices)
        out_lookup: dict[tuple[Any, ...], _V] = {}
        i = 0
        for row in labels:
            row_filt = tuple(row[i] for i in indices)
            if row_filt not in out_lookup:
                out_lookup[row_filt] = self.values[i % size]
                i += 1

        ret = [out_lookup[tuple(_r[i] for i in indices)] for _r in labels]
        return ret

    def map(
        self,
        values: DataFrameWrapper,  # the data frame
    ) -> Sequence[_V]:
        if self._by:
            series = [values[k] for k in self._by]
        else:
            # constant, no key filter
            return self.values[0]
        uniques = [unique(ar, axis=None) for ar in series]
        out = np.empty(series[0].shape, dtype=object)
        i = 0
        for row in itertools.product(*uniques):
            sl: NDArray[np.bool_] = np.all(
                np.column_stack([a == b for a, b in zip(series, row)]), axis=1
            )
            ntrue = sl.sum()
            if ntrue > 0:
                val = self.values[i % len(self.values)]
                out[sl] = np.full(ntrue, val, dtype=object)
                i += 1
        return out


class MapPlan(ABC, Generic[_V]):
    def __init__(
        self,
        on: tuple[str, ...],
        mapper: Callable[[dict[str, np.ndarray]], Sequence[_V]],
    ):
        self._on = on
        self._mapper = mapper

    def __repr__(self) -> str:
        cname = type(self).__name__
        if self._on:
            return f"{cname}(on={self._on!r}, mapper={self._mapper})"
        else:
            return f"{cname}(mapper={self._mapper!r})"

    @classmethod
    def from_map(
        cls,
        on: Sequence[str],
        mapper: Callable[[dict[str, np.ndarray]], Sequence[_V]],
    ) -> Self:
        if not isinstance(on, (tuple, list)):
            raise TypeError(f"on must be a sequence, not {type(on)}")
        return cls(tuple(on), mapper)

    @classmethod
    def from_const(cls, value: _V) -> Self:
        """Create a map plan that always returns a constant value."""
        return cls.from_map((), ConstMap(value))

    def is_const(self) -> bool:
        """Return True if the plan is a constant plan."""
        return isinstance(self._mapper, ConstMap)

    @classmethod
    def default(cls) -> Self:
        return cls((), cls._default_mapper())

    @classmethod
    @abstractmethod
    def _default_mapper(cls) -> Callable[[dict[str, np.ndarray]], Sequence[_V]]:
        """Default mapper for the plan."""

    def map(
        self,
        values: DataFrameWrapper,  # the data frame
    ) -> Sequence[_V]:
        """Calculate the values for the input data frame."""
        if self._on:
            input_dict = {k: values[k] for k in self._on}
        else:
            # constant or default size, no key filter
            try:
                k, v = next(iter(values.iter_items()))
                input_dict = {k: v}
            except StopIteration:
                input_dict = {}
        return self._mapper(input_dict)


class ScalarMapPlan(MapPlan[float]):
    @classmethod
    def from_range(cls, on: str, limits: tuple[float, float] | None = None) -> Self:
        """Add a mapper that maps a range to a value."""

        def mapper(values: dict[str, np.ndarray]) -> Sequence[_V]:
            arr = values[on]
            valid = np.isfinite(arr)
            amin, amax = arr[valid].min(), arr[valid].max()
            if amin == amax:
                if limits is not None:
                    w0, w1 = limits
                    return np.full(arr.shape, (w0 + w1) / 2)
                else:
                    return np.full(arr.shape, amin)
            _arr = arr.clip(amin, amax)
            _arr[np.isnan(_arr)] = amin
            if limits is not None:
                w0, w1 = limits
                return (_arr - amin) / (amax - amin) * (w1 - w0) + w0
            else:
                return _arr

        return cls.from_map((on,), mapper)


class ColorPlan(CyclicPlan[Color]):
    @classmethod
    def _default_values(cls) -> list[Color]:
        return [Color("black")]  # NOTE: will be updated by the canvas palette

    @classmethod
    def _norm_value(cls, v: Any) -> Color:
        return Color(v)

    @classmethod
    def from_palette(
        cls,
        by: Sequence[str],
        palette: ColorPalette | None = None,
    ) -> ColorPlan:
        if palette is None:
            palette = "tab10"
        palette = ColorPalette(palette)
        colors = palette.nextn(palette.ncolors, update=False)
        return cls(tuple(by), colors)

    # NOTE: Color instance is detected as a sequence of 4 floats
    #       so we need to override the default mapper
    def map(
        self,
        values: dict[str, np.ndarray],  # the data frame
    ) -> Sequence[_V]:
        if self._by:
            series = [values[k] for k in self._by]
        else:
            # constant, no key filter
            return self.values[0]
        uniques = [unique(ar, axis=None) for ar in series]
        out = np.empty((series[0].size, 4), dtype=np.float32)
        i = 0
        for row in itertools.product(*uniques):
            sl: NDArray[np.bool_] = np.all(
                np.column_stack([a == b for a, b in zip(series, row)]), axis=1
            )
            ntrue = sl.sum()
            if ntrue > 0:
                val = self.values[i % len(self.values)]
                out[sl] = [val] * ntrue
                i += 1
        return out


class StylePlan(CyclicPlan[LineStyle]):
    @classmethod
    def _default_values(cls) -> list[LineStyle]:
        return [
            LineStyle.SOLID,
            LineStyle.DASH,
            LineStyle.DOT,
            LineStyle.DASH_DOT,
        ]

    @classmethod
    def _norm_value(cls, v: Any) -> LineStyle:
        return LineStyle(v)

    def with_choices(self, choices: Sequence[LineStyle]) -> ColorPlan:
        choices = [LineStyle(s) for s in choices]
        return ColorPlan(self._by, choices)


class HatchPlan(CyclicPlan[Hatch]):
    @classmethod
    def _default_values(cls) -> list[Hatch]:
        return [
            Hatch.SOLID,
            Hatch.DIAGONAL_BACK,
            Hatch.HORIZONTAL,
            Hatch.CROSS,
            Hatch.VERTICAL,
            Hatch.DIAGONAL_CROSS,
            Hatch.DOTS,
            Hatch.DIAGONAL_FORWARD,
        ]

    @classmethod
    def _norm_value(cls, v: Any) -> Hatch:
        return Hatch(v)


class SymbolPlan(CyclicPlan[Symbol]):
    @classmethod
    def _default_values(cls) -> list[Symbol]:
        return [
            Symbol.CIRCLE,
            Symbol.SQUARE,
            Symbol.TRIANGLE_UP,
            Symbol.DIAMOND,
            Symbol.TRIANGLE_DOWN,
        ]

    @classmethod
    def _norm_value(cls, v: Any) -> Symbol:
        return Symbol(v)


class ColormapPlan(MapPlan[NDArray[np.float32]]):
    @classmethod
    def _default_mapper(cls):
        return lambda x: np.ones((x.size, 4))

    @classmethod
    def from_colormap(
        cls,
        on: str,
        cmap: Colormap,
        *,
        clim: tuple[float, float] | None = None,
    ) -> Self:
        def mapper(values: dict[str, np.ndarray]) -> Sequence[_V]:
            arr = values[on]
            valid = np.isfinite(arr)
            if clim is None:
                amin, amax = arr[valid].min(), arr[valid].max()
            else:
                amin, amax = clim
            if amin == amax:
                color = cmap(0.5)
                return np.full((arr.size, 4), np.asarray(color))
            _arr = arr.clip(amin, amax)
            _arr[np.isnan(_arr)] = amin
            return cmap((_arr - amin) / (amax - amin))

        return cls.from_map((on,), mapper)


class WidthPlan(ScalarMapPlan):
    @classmethod
    def _default_mapper(cls):
        return lambda _: 1.0


class SizePlan(ScalarMapPlan):
    @classmethod
    def _default_mapper(cls):
        return lambda _: 12.0


def _to_intervals(each_uniques: list[np.ndarray]):
    each_size = [un.size for un in each_uniques] + [1]
    return np.cumprod(each_size[1:][::-1])[::-1]


class ConstMap:
    def __init__(self, value):
        self._value = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}<{self._value!r}>"

    def __call__(self, values: dict[str, np.ndarray]) -> Sequence[_V]:
        series = next(iter(values.values()), np.zeros(0))
        return [self._value] * series.size



import numpy as np

from whitecanvas import new_canvas
from ._utils import assert_color_array_equal
import pytest

def test_cat(backend: str):
    canvas = new_canvas(backend=backend)
    rng = np.random.default_rng(1642)
    df = {
        "x": rng.normal(size=30),
        "y": rng.normal(size=30),
        "label": np.repeat(["A", "B", "C"], 10),
    }
    canvas.cat(df, "x", "y").add_line()
    canvas.cat(df, "x", "y").add_line(color="label")
    canvas.cat(df, "x", "y").add_markers()
    canvas.cat(df, "x", "y").add_markers(color="label")
    canvas.cat(df, "x", "y").add_markers(hatch="label")
    canvas.cat(df, "x", "y").add_hist2d(bins=5)
    canvas.cat(df, "x", "y").add_hist2d(bins=(5, 4))
    canvas.cat(df, "x", "y").add_hist2d(bins="auto")
    canvas.cat(df, "x", "y").add_hist2d(bins=("auto", 5))
    canvas.cat(df, "x", "y").along_x().add_hist(bins=5)
    canvas.cat(df, "x", "y").along_x().add_hist(bins=5, color="label")
    canvas.cat(df, "x", "y").along_y().add_hist(bins=6)
    canvas.cat(df, "x", "y").along_y().add_hist(bins=6, color="label")

@pytest.mark.parametrize("orient", ["v", "h"])
def test_cat_plots(backend: str, orient: str):
    canvas = new_canvas(backend=backend)
    df = {
        "y": np.arange(30),
        "label": np.repeat(["A", "B", "C"], 10),
        "c": ["P", "Q"] * 15,
    }
    if orient == "v":
        cat_plt = canvas.cat_x(df, "label", "y")
    else:
        cat_plt = canvas.cat_y(df, "y", "label")
    cat_plt.add_stripplot(color="c")
    cat_plt.add_swarmplot(color="c")
    cat_plt.add_boxplot(color="c")
    cat_plt.add_violinplot(color="c")

def test_colored_plots(backend: str):
    canvas = new_canvas(backend=backend)
    df = {
        "x": np.arange(30),
        "y": np.arange(30),
        "label": np.repeat(["A", "B", "C"], 10),
    }

    canvas.cat(df, "x", "y").add_markers(color="label")
    canvas.cat(df, "x", "y").add_line(color="label")

def test_markers(backend: str):
    canvas = new_canvas(backend=backend)
    df = {
        "x": np.arange(30),
        "y": np.arange(30),
        "size": np.arange(30) / 2 + 8,
        "label0": np.repeat(["A", "B", "C"], 10),  # [A, A, ..., B, B, ..., C, C, ...]
        "label1": ["One"] * 10 + ["Two"] * 20,
    }

    _c = canvas.cat(df, "x", "y")
    out = _c.add_markers(color="label0", size="size", symbol="label1")
    assert len(set(out._base_layer.symbol[:10])) == 1
    assert len(set(out._base_layer.symbol[10:])) == 1

    out = _c.add_markers(color="label1", size="size", hatch="label0")
    assert len(set(out._base_layer.face.hatch[:10])) == 1
    assert len(set(out._base_layer.face.hatch[10:20])) == 1
    assert len(set(out._base_layer.face.hatch[20:])) == 1

    out = _c.add_markers(color="label1").with_edge(color="label0")
    assert len(np.unique(out._base_layer.edge.color[:10], axis=0)) == 1
    assert len(np.unique(out._base_layer.edge.color[10:20], axis=0)) == 1
    assert len(np.unique(out._base_layer.edge.color[20:], axis=0)) == 1

    # test scalar color
    out = _c.add_markers(color="black")
    assert_color_array_equal(out._base_layer.face.color, "black")

    out = _c.add_markers(color="transparent").with_edge_colormap("size")

def test_heatmap(backend: str):
    canvas = new_canvas(backend=backend)
    df = {
        "x": ["A", "B", "A", "B", "A", "B"],
        "y": ["P", "P", "Q", "Q", "R", "R"],
        "z": [1, 2, 3, 4, 5, 6],
    }
    im = canvas.cat_xy(df, "x", "y").first().add_heatmap(value="z")
    canvas.imref(im).add_text()

    df = {
        "x": ["A", "A", "A", "B", "A", "B"],
        "y": ["P", "Q", "Q", "Q", "P", "Q"],
        "z": [1.1, 2.1, 3.4, 6.4, 1.1, 6.8],
    }
    im = canvas.cat_xy(df, "x", "y").mean().add_heatmap(value="z", fill=-1)
    canvas.imref(im).add_text(fmt=".1f")
    assert im.clim == (1.1, 6.6)



from __future__ import annotations

import warnings

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend.plotly._base import (
    PlotlyLayer,
    from_plotly_linestyle,
    to_plotly_linestyle,
)
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.types import LineStyle
from whitecanvas.utils.normalize import arr_color, rgba_str_color
from whitecanvas.utils.type_check import is_real_number


@check_protocol(LineProtocol)
class MonoLine(PlotlyLayer):
    def __init__(self, xdata, ydata):
        self._props = {
            "x": xdata,
            "y": ydata,
            "mode": "lines",
            "line": {"color": "blue", "width": 1, "dash": "solid", "simplify": False},
            "type": "scatter",
            "showlegend": False,
            "visible": True,
        }

    def _plt_get_data(self):
        return self._props["x"], self._props["y"]

    def _plt_set_data(self, xdata, ydata):
        self._props["x"] = xdata
        self._props["y"] = ydata

    def _plt_get_edge_width(self) -> float:
        return self._props["line"]["width"]

    def _plt_set_edge_width(self, width: float):
        self._props["line"]["width"] = width

    def _plt_get_edge_style(self) -> LineStyle:
        return from_plotly_linestyle(self._props["line"]["dash"])

    def _plt_set_edge_style(self, style: LineStyle):
        self._props["line"]["dash"] = to_plotly_linestyle(style)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return arr_color(self._props["line"]["color"])

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self._props["line"]["color"] = rgba_str_color(color)

    def _plt_get_antialias(self) -> bool:
        return not self._props["line"]["simplify"]

    def _plt_set_antialias(self, antialias: bool):
        self._props["line"]["simplify"] = not antialias


@check_protocol(MultiLineProtocol)
class MultiLine(PlotlyLayer):
    def __init__(self, data: list[NDArray[np.floating]]):
        # In plotly, we can break a line into multiple segments by inserting None
        xdata = []
        ydata = []
        for each in data:
            xdata = xdata + each[:, 0].tolist() + [None]
            ydata = ydata + each[:, 1].tolist() + [None]
        # remove last None
        if len(xdata) > 0:
            xdata.pop()
            ydata.pop()
        self._data = data

        self._props = {
            "x": xdata,
            "y": ydata,
            "mode": "lines",
            "line": {
                "color": "blue",
                "width": 1,
                "dash": "solid",
                "simplify": False,
            },
            "type": "scatter",
            "showlegend": False,
            "visible": True,
        }

    def _plt_get_data(self):
        return self._data

    def _plt_set_data(self, data: list[NDArray[np.floating]]):
        xdata = []
        ydata = []
        for each in data:
            xdata = xdata + each[:, 0].tolist() + [None]
            ydata = ydata + each[:, 1].tolist() + [None]
        # remove last None
        if len(xdata) > 0:
            xdata.pop()
            ydata.pop()
        self._data = data
        self._props["x"] = xdata
        self._props["y"] = ydata

    def _plt_get_edge_width(self) -> NDArray[np.float32]:
        width = self._props["line"]["width"]
        return np.full(len(self._data), width, dtype=np.float32)

    def _plt_set_edge_width(self, width):
        if is_real_number(width):
            w = width
        else:
            candidates = np.unique(width)
            if len(candidates) == 1:
                w = candidates[0]
            elif len(candidates) == 0:
                w = 0.0
            else:
                warnings.warn(
                    "plotly does not support multiple line widths. "
                    "Set to the first one.",
                    UserWarning,
                    stacklevel=2,
                )
                w = width[0]
        self._props["line"]["width"] = w

    def _plt_get_edge_style(self) -> list[LineStyle]:
        style = from_plotly_linestyle(self._props["line"]["dash"])
        return [style] * len(self._data)

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            self._props["line"]["dash"] = to_plotly_linestyle(style)
        else:
            candidates = set(style)
            if len(candidates) == 1:
                self._props["line"]["dash"] = to_plotly_linestyle(style[0])
            elif len(candidates) == 0:
                self._props["line"]["dash"] = "solid"
            else:
                warnings.warn(
                    "plotly does not support multiple line styles. "
                    "Set to the first one.",
                    UserWarning,
                    stacklevel=2,
                )
                self._props["line"]["dash"] = to_plotly_linestyle(style[0])

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        col = arr_color(self._props["line"]["color"])
        return np.stack([col] * len(self._data), axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            self._props["line"]["color"] = rgba_str_color(color)
        else:
            candidates = np.unique(color, axis=0)
            if len(candidates) == 1:
                self._props["line"]["color"] = rgba_str_color(color[0])
            elif len(candidates) == 0:
                self._props["line"]["color"] = "blue"
            else:
                warnings.warn(
                    "plotly does not support multiple line colors. "
                    "Set to the first one.",
                    UserWarning,
                    stacklevel=2,
                )
                self._props["line"]["color"] = rgba_str_color(color[0])

    def _plt_get_antialias(self) -> bool:
        return not self._props["line"]["simplify"]

    def _plt_set_antialias(self, antialias: bool):
        self._props["line"]["simplify"] = not antialias
